class UserError(Exception):
    pass


class AmbiguousPathError(UserError):
    pass


class CardinalityError(UserError):
    pass


class AttributeNameError(UserError):
    pass


class DisjointPathError(UserError):
    pass



from typing import Callable, Union, Tuple, Dict, Any, List
from math import floor, ceil
import functools

import numpy as np

from .utilities import mask_infs
from .objects import AttributeQuery, ObjectQuery
from .base import BaseQuery


def _template_operator(string_op: str, name: str, item: BaseQuery, python_func: Callable = None,
                       remove_infs=True, in_dtype=None, out_dtype=None, *args, **kwargs):
    if not isinstance(item, AttributeQuery):
        if python_func is None:
            raise NotImplementedError(f"{name} is not implemented for {type(item)}")
        return python_func(item, *args, **kwargs)
    if remove_infs:
        string_op = string_op.format(mask_infs('{0}'))
    return item._perform_arithmetic(string_op, name, expected_dtype=in_dtype, returns_dtype=out_dtype)


def sign(item, *args, **kwargs):
    return _template_operator('sign({0})', 'sign', item, np.sign, remove_infs=True, out_dtype='float', args=args, kwargs=kwargs)


def exp(item, *args, **kwargs):
    return _template_operator('exp({0})', 'exp', item, np.exp, remove_infs=True, out_dtype='float', args=args, kwargs=kwargs)


def log(item, *args, **kwargs):
    return _template_operator('log({0})', 'log', item, np.log, remove_infs=True, out_dtype='float', args=args, kwargs=kwargs)


def log10(item, *args, **kwargs):
    return _template_operator('log10({0})', 'log10', item, np.log10, remove_infs=True, out_dtype='float', args=args, kwargs=kwargs)


def sqrt(item, *args, **kwargs):
    return _template_operator('sqrt({0})', 'sqrt', item, np.sqrt, remove_infs=True, out_dtype='float', args=args, kwargs=kwargs)

def ismissing(item):
    return _template_operator('{0} is null', 'isnull', item, lambda x: x is None, remove_infs=True, out_dtype='boolean')
isnull = ismissing

def isnan(item):
    return _template_operator('{0} == 1.0/0.0', 'isnan', item, np.isnan, remove_infs=False, out_dtype='boolean')

def _object_scalar_operator(item: ObjectQuery, op_string: str, op_name: str, returns_type: str):
    n, wrt = item._G.add_scalar_operation(item._node, op_string, op_name, parameters=None)
    return AttributeQuery._spawn(item, n, index_node=wrt, single=True, dtype=returns_type, factor_name=op_name)

def neo4j_id(item: ObjectQuery):
    return _object_scalar_operator(item, 'id({0})', 'neo4j_id', 'number')

def reduce(function, sequence: List, initial=None):
    if not len(sequence):
        return initial
    if initial is None and len(sequence) == 1:
        return sequence[0]
    if any(isinstance(s, AttributeQuery) for s in sequence):
        if initial is None:
            initial = sequence[0]
            sequence = sequence[1:]
        if not isinstance(initial, AttributeQuery):
            raise TypeError(f"The initial item must be an AttributeQuery")
        _list = ', '.join([f'{{{i+1}}}' for i in range(len(sequence))])
        i = initial._G.get_variable_name('i')
        acc = initial._G.get_variable_name('acc')
        test_query = function(initial, sequence[0])  # Type: AttributeQuery
        if len(sequence) == 1:
            return test_query
        returns_dtype = test_query.dtype
        string_op = test_query._G.G.nodes[test_query._node]['variables'][0]
        string_op = string_op.replace('{0}', f"{acc}").replace('{1}', f"{i}")
        string_op = string_op.replace(test_query._G.G.nodes[initial._node]['variables'][0], f"{acc}")
        string_op = string_op.replace(test_query._G.G.nodes[sequence[0]._node]['variables'][0], f"{i}")
        return initial._perform_arithmetic(f"reduce({acc} = {{0}}, {i} IN [{_list}] | {string_op})", 'reduce',
                                           *sequence, returns_dtype=returns_dtype)
    if initial:
        return functools.reduce(function, sequence, initial)
    return functools.reduce(function, sequence)

def switch(item: AttributeQuery, states: Union[Dict[Any, Union[AttributeQuery, Any]], Union[Any, AttributeQuery]],
           _else: Union[Any, AttributeQuery] = None):
    """
    Perform a CASE/SWITCH operation with `item` as the variable to test and with `states` being either:
        - dict of {state: result}
        - result
    So there are two call signatures
    for example:
        >>> switch(query > 1, yes, no)
        >>> switch(query, {1: querya, 2: queryb})
    """
    if not isinstance(item, AttributeQuery):
        raise TypeError(f"`item` must be an AttributeQuery")
    if not isinstance(states, dict):
        states = {True: states}

    dtypes = {result.dtype if isinstance(result, AttributeQuery) else type(result) for result in list(states.values())+[_else]}
    if ObjectQuery in dtypes:
        raise TypeError(f"Switching on objects is not yet supported")
    if dtypes.issubset({'integer', int, np.int, np.int_, None, type(None)}):
        dtype = 'integer'
    elif dtypes.issubset({'float', float, np.float, np.float_, None, type(None)}):
        dtype = 'float'
    elif dtypes.issubset({'number', float, int, None, type(None)}):
        dtype = 'number'
    elif dtypes.issubset({'boolean', bool, None, type(None)}):
        dtype = 'boolean'
    elif dtypes.issubset({'string', str, None, type(None)}):
        dtype = 'string'
    else:
        raise TypeError(f"All switch result values, including else, must be of the same type. Types given: {dtypes}")
    params = []
    vars = []
    state_list = []
    for i, (k, v) in enumerate(states.items()):
        i *= 2
        kn = f'{{{i}}}'
        vn = f"{{{i+1}}}"
        if not isinstance(k, BaseQuery):
            k = item._G.add_parameter(k)
            params.append(k)
            kn = k
        if not isinstance(v, BaseQuery):
            v = item._G.add_parameter(v)
            params.append(v)
            vn = v
        _switch = f'{kn} THEN {vn}'  # k then v
        state_list.append(_switch)
        vars.append(k)
        vars.append(v)
    if _else is None:
        _else = 'null'
    elif not isinstance(_else, BaseQuery):
        _else = item._G.add_parameter(_else)
        params.append(_else)
        vars.append(_else)

    nodes = [v._node for v in vars if isinstance(v, BaseQuery)]
    switches = ' WHEN '.join(state_list)
    statement = f'CASE {{0}} WHEN {switches} ELSE {_else} END'
    if not nodes:
        n, wrt = item._G.add_scalar_operation(item._node, statement, 'op-switch', parameters=params)
    else:
        n, wrt = item._G.add_combining_operation(statement, 'op-switch', item._node, *nodes, parameters=params)
    return AttributeQuery._spawn(item, n, index_node=wrt, single=True, dtype=dtype, factor_name='switch')

def to_int(x):
    return _template_operator('tointeger({0})', 'int', x, int, remove_infs=True, out_dtype='float')


import logging
from collections.abc import Iterable, Mapping
from datetime import datetime
from time import sleep
from warnings import warn

import py2neo
from astropy.table import Table
from pathlib import Path
from py2neo import Graph as NeoGraph

import numpy as np
import pandas as pd
from weaveio.context import ContextMeta
from weaveio.writequery import CypherQuery

missing_types = {int:  np.inf, float: np.inf, str: '<MISSING>', type(None): np.inf, None: np.inf, bool: np.inf, datetime: datetime(1900, 1, 1, 0, 0)}
convert_types = {bool: bool, np.bool: bool, np.bool_: bool, int:  int, float: float, str: str, type(None): float, None: float,
                 datetime: lambda x: datetime(x.year, x.month, x.day, x.hour, x.minute, x.second),
                 list: list, tuple: tuple, np.ndarray: np.ndarray, pd.DataFrame: pd.DataFrame, pd.Series: pd.Series,
                 dict: dict, Table: Table,
                 np.int64: int, np.float64: float, np.float32: float, np.int32:int,
                 np.str_: str, np.str: str,
                 Path: str}  # order matters

def is_null(x):
    try:
        return np.all(x != x)
    except TypeError:
        pass
    return x is None

def stringify(x):
    if isinstance(x, Mapping):
        vars = ', '.join([f"{k}:{stringify(v)}" if isinstance(k, str) else f"{stringify(k)}:{stringify(v)}" for k, v in x.items()])
        return f"{{{vars}}}"
    elif isinstance(x, Iterable) and not isinstance(x, str):
        return "[" + ', '.join([stringify(i) for i in x]) + "]"
    elif isinstance(x, str):
        return f'"{x}"'
    else:
        return str(x)

def _convert_datatypes(x, nan2missing=True, none2missing=True, surrounding_type=None):
    if isinstance(x, (tuple, list)):
        types = set(convert_types[type(i)] for i in x if not is_null(i))
        if len(types) == 0:
            acceptable_type = str
        elif len(types) == 1:
            acceptable_type = types.pop()
        else:
            raise TypeError(f"Lists must be of homogeneous type")
        r = [_convert_datatypes(xi, nan2missing, none2missing, surrounding_type=acceptable_type) for xi in x]
        if isinstance(x, tuple):
            r = tuple(r)
        return r
    elif isinstance(x, dict):
        return {str(_convert_datatypes(k, nan2missing, none2missing)): _convert_datatypes(v, nan2missing, none2missing) for k, v in x.items()}
    elif isinstance(x, (pd.DataFrame, pd.Series)):
        return _convert_datatypes(pd.DataFrame(x).reset_index().to_dict('records'), nan2missing, none2missing)
    elif isinstance(x, Table):
        x = list(map(lambda row: {c: ri for c, ri in zip(x.colnames, row)}, x.iterrows()))
        return _convert_datatypes(pd.DataFrame(x).reset_index().to_dict('records'), nan2missing, none2missing)
    elif isinstance(x, np.ndarray):
        return _convert_datatypes(x.tolist(), nan2missing, none2missing)
    if not (none2missing or nan2missing):
        return x
    elif none2missing:
        if x is None:
            return missing_types[surrounding_type]
    if nan2missing:
        try:
            if np.isnan(x):
                return missing_types[surrounding_type]
        except TypeError:
            pass
    if surrounding_type is not None:
        return convert_types[surrounding_type](x)
    for from_type, to_type in convert_types.items():
        if from_type is not None:
            if isinstance(x, from_type):
                x = to_type(x)
                break
    return x


class Graph(metaclass=ContextMeta):
    def __new__(cls, *args, **kwargs):
        # resolves the parent instance
        instance = super().__new__(cls)
        if kwargs.get("graph") is not None:
            instance._parent = kwargs.get("graph")
        else:
            instance._parent = cls.get_context(error_if_none=False)
        return instance

    def __init__(self, profile=None, name=None, **settings):
        self.write_allowed = settings.pop('write', False)
        self.neograph = NeoGraph(profile, name, **settings)

    def create_unique_constraint(self, label, key):
        try:
            self.neograph.schema.create_uniqueness_constraint(label, key)
        except py2neo.database.work.ClientError:
            pass

    def drop_unique_constraint(self, label, key):
        try:
            self.neograph.schema.drop_uniqueness_constraint(label, key)
        except py2neo.database.work.DatabaseError:
            pass

    def write(self, collision_manager):
        return CypherQuery(collision_manager)

    def _execute(self, cypher, parameters, backoff=1, limit=10):
        if not isinstance(cypher, str):
            raise TypeError(f"Cypher must be a string")
        return self.neograph.auto(readonly=not self.write_allowed).run(cypher, parameters=parameters)

    def execute(self, cypher, **payload):
        d = _convert_datatypes(payload, nan2missing=True, none2missing=True)
        try:
            return self._execute(cypher, d)
        except IndexError:
            raise ConnectionResetError(f"Py2neo dropped the connection because it was taking too long. "
                                       f"Split up your query using batch_size=??")

    def output_for_debug(self,  arrow=False, cmdline=False, silent=False, **payload):
        d = _convert_datatypes(payload, nan2missing=True, none2missing=True)
        if not silent:
            warn(f"When parameters are output for debug in the neo4j desktop, it cannot be guaranteed the data types will remain the same. "
                 f"For certain, infs/nans/None are converted to strings (to avoid this, run your query without using `output_for_debug`)")
        if arrow:
            if cmdline:
                return ' '.join([f'-P "{k} => {stringify(v)}"' for k, v in d.items()])
            return '\n'.join([f':param {k} => {stringify(v)}' for k, v in d.items()])
        return f':params {d}'

Graph._context_class = Graph



"""
Hierarchies are defined like so:

class Name(Type):
    singular_name = 'custom_singular_name'
    plural_name = 'custom_plural_name'
    idname = 'the_name_of_the_unique_attribute_of_this_Object'
    identifier_builder = [list_of_attributes_or_objects_that_define_uniqueness]
    factors = ['attribute1', 'attribute2']
    products = ['name_of_product']
    parents = [required_object_that_comes_before]
    children = [required_object_that_comes_after]

if an object of type O requires n parents of type P then this is equivalent to defining that instances of those behave as:
    parent-(n)->object (1 object has n parents of type P)
    it implicitly follows that:
        object--(m)-parent (each of object's parents of type P can be used by an unknown number `m` of objects of type O = many to one)
if an object of type O requires n children of type C then this is equivalent to defining that instances of those behave as:
    object-(n)->child (1 object has n children of type C)
    it implicitly follows that:
        child-[has 1]->Object (each child has maybe 1 parent of type O)

A child does not need to require a parent at instantiation

You can modify the first relation in each of the above by using:
    Multiple(parent/child, minnumber, maxnumber)
    One2One(parent/child) - makes a reciprocal one to one relationship
    Optional(parent/child) == Multiple(parent/Child, 0, 1)

"""
import inspect
import sys

import numpy as np

from weaveio.config_tables import progtemp_config
from weaveio.hierarchy import Hierarchy, Multiple, OneOf, Optional


class ArrayHolder(Hierarchy):
    is_template = True


class WavelengthHolder(ArrayHolder):
    singular_name = 'wavelength_holder'
    factors = ['wvl', 'cd1_1', 'crval1', 'naxis1']
    identifier_builder = ['cd1_1', 'crval1', 'naxis1']


class BaseMeasurement(Hierarchy):
    factors = ['value', 'err']
    indexes = ['value']
    is_template = True


class Measurement(BaseMeasurement):
    pass


class Magnitude(BaseMeasurement):
    pass


class Author(Hierarchy):
    is_template = True


class CASU(Author):
    """
    CASU is the pipeline that will produce the L1 and Raw data files and spectra.
    The version of CASU that produces these files may change without warning and files may be duplicated.
    This CASU object breaks that potential degeneracy.
    """
    idname = 'id'


class APS(Author):
    """
    APS is the pipeline that will produce the L2 data files and model spectra.
    The version of APS that produces these files may change without warning and files may be duplicated.
    This APS object breaks that potential degeneracy.
    """
    idname = 'version'


class System(Author):
    idname = 'version'


class ArmConfig(Hierarchy):
    """
    An ArmConfig is the entire configuration of one arm of the WEAVE spectrograph.
    The ArmConfig is therefore a subset of the ProgTemp code and there are multiple ways to identify
    one:
    - resolution can be "high" or "low"
    - vph is the number of the grating [1=lowres, 2=highres, 3=highres(green)]
    - camera can be 'red' or 'blue'
    - colour can be 'red', 'blue', or 'green'
    """
    factors = ['resolution', 'vph', 'camera', 'colour', 'colour_code']
    identifier_builder = ['resolution', 'vph', 'camera']

    def __init__(self, tables=None, **kwargs):
        if not kwargs.get('do_not_create', False):
            if kwargs['vph'] == 3 and kwargs['camera'] == 'blue':
                kwargs['colour'] = 'green'
            else:
                kwargs['colour'] = kwargs['camera']
            kwargs['colour_code'] = kwargs['colour'][0].upper()
        super().__init__(tables=tables, **kwargs)

    @classmethod
    def from_progtemp_code(cls, progtemp_code):
        config = progtemp_config.loc[progtemp_code[0]]
        red = cls(resolution=str(config.resolution), vph=int(config.red_vph), camera='red')
        blue = cls(resolution=str(config.resolution), vph=int(config.blue_vph), camera='blue')
        return red, blue


class Obstemp(Hierarchy):
    """
    Whilst ProgTemp deals with "how" a target is observed, OBSTEMP deals with "when" a target is observed,
    namely setting the observational constraints required to optimally extract scientific information from the observation.
    The ProgTemp is made up of maxseeing, mintrans (transparency), minelev (elevation), minmoon, minsky and
    each combination is given a valid code, where each letter corresponds to one of those settings.
    """
    factors = [
        'maxseeing', 'mintrans', 'minelev', 'minmoon', 'minsky', 'maxairmass',
        'maxseeing_grade', 'mintrans_grade', 'minelev_grade', 'minmoon_grade', 'minsky_grade',
    ]
    idname = 'code'
    seeings = {'A': 0.7, 'B': 0.8, 'C': 0.9, 'D': 1.0, 'E': 1.1, 'F': 1.2, 'G': 1.3, 'H': 1.4, 'I': 1.5,
               'J': 1.6, 'K': 1.7, 'L': 1.8, 'M': 1.9, 'N': 2.0, 'O': 2.1, 'P': 2.2, 'Q': 2.3, 'R': 2.4,
               'S': 2.5, 'T': 2.6, 'U': 2.7, 'V': 2.8, 'W': 2.9, 'X': 3.0}
    transparencies = {'A': 0.8, 'B': 0.7, 'C': 0.6, 'D': 0.5, 'E': 0.4}
    elevs = {'A': 50.28, 'B': 45.58, 'C': 41.81, 'D': 35.68, 'E': 33.75, 'F': 25.00}
    airmasses = {'A': 1.3, 'B': 1.4, 'C': 1.5, 'D': 1.6, 'E': 1.8, 'F': 2.4}
    moons = {'A': 90, 'B': 70, 'C': 50, 'D': 30, 'E': 0}
    brightnesses = {'A': 21.7, 'B': 21.5, 'C': 21.0, 'D': 20.5, 'E': 19.6, 'F': 18.5, 'G': 17.7}

    @classmethod
    def from_header(cls, header):
        names = [f.lower() for f in cls.factors[:-1]]
        obstemp_code = list(header['OBSTEMP'])
        data = {n+'_grade': v for v, n in zip(obstemp_code, names)}
        data['maxseeing'] = cls.seeings[data['maxseeing_grade']]
        data['mintrans'] = cls.transparencies[data['mintrans_grade']]
        data['minelev'] = cls.elevs[data['minelev_grade']]
        data['maxairmass'] = cls.airmasses[data['minelev_grade']]
        data['minmoon'] = cls.moons[data['minmoon_grade']]
        data['minsky'] = cls.brightnesses[data['minsky_grade']]
        return cls(code=header['OBSTEMP'], **data)


class Survey(Hierarchy):
    """
    A Survey is one of the official Surveys of WEAVE. WL is one of them.
    """
    idname = 'name'


class WeaveTarget(Hierarchy):
    """
    A WeaveTarget is the disambiguated target to which all targets submitted by surveys belong.
    The "cname" of a weavetarget will be formed of the ra and dec of the submitted target.
    """
    idname = 'cname'


class Fibre(Hierarchy):
    """
    A WEAVE spectrograph fibre
    """
    idname = 'id'


class Targprog(Hierarchy):
    """
    Sub-programme name within a survey (TARGPROG)
    """
    singular_name = 'targprog'
    parents = [Survey]
    factors = ['name']
    identifier_builder = ['name', 'survey']


class Catalogue(Hierarchy):
    """
    A catalogue which was submitted by a survey.
    """
    parents = [Survey, Multiple(Targprog, maxnumber=600)]
    factors = ['name']
    identifier_builder = ['name', 'survey']


class SurveyTarget(Hierarchy):
    """
    A target which was submitted by a subprogramme contained within a catalogue. This is likely
    the target you want if you not linking observations between subprogrammes.
    targname is optional for MOS observations
    """
    parents = [Catalogue, WeaveTarget, Multiple(Targprog, maxnumber=5)]
    factors = ['targid', 'targname', 'targra', 'targdec', 'epoch', 'targuse',
               'targclass', 'targpmra', 'targpdec', 'targparal', 'targprio'] \
              + Magnitude.as_factors('g', 'r', 'i', 'gg', 'bp', 'rp', prefix='mag_')
    identifier_builder = ['weave_target', 'catalogue', 'targid', 'targra', 'targdec', 'targuse']


class InstrumentConfiguration(Hierarchy):
    """
    The WEAVE instrument can be configured into MOS/LIFU/mIFU modes and the spectral binning in pixels.
    InstrumentConfiguration holds the mode, binning, and is linked to an ArmConfig.
    """
    factors = ['mode', 'binning']
    parents = [Multiple(ArmConfig, 2, 2)]
    identifier_builder = ['arm_configs', 'mode', 'binning']


class Progtemp(Hierarchy):
    """
    The ProgTemp code is an integral part of describing a WEAVE target.
    This parameter encodes the requested instrument configuration, OB length, exposure time,
    spectral binning, cloning requirements and probabilistic connection between these clones.
    The ProgTemp is therefore formed from instrumentconfiguration, length, and an exposure_code.
    """
    parents = [InstrumentConfiguration]
    factors = ['length', 'exposure_code', 'code']
    identifier_builder = ['instrument_configuration'] + factors

    @classmethod
    def from_progtemp_code(cls, progtemp_code):
        progtemp_code = progtemp_code.split('.')[0]
        progtemp_code_list = list(map(int, progtemp_code))
        configs = ArmConfig.from_progtemp_code(progtemp_code_list)
        mode = progtemp_config.loc[progtemp_code_list[0]]['mode']
        binning = progtemp_code_list[3]
        config = InstrumentConfiguration(arm_configs=configs, mode=mode, binning=binning)
        exposure_code = progtemp_code[2:4]
        length = progtemp_code_list[1]
        return cls(code=progtemp_code, length=length, exposure_code=exposure_code,
                   instrument_configuration=config)


class OBSpec(Hierarchy):
    """
    When an xml observation specification is submitted to WEAVE, an OBSpec is created containing all
    the information about when and how to observe.
    When actually observing them, an "OB" is created with its own unique obid.
    """
    singular_name = 'obspec'
    factors = ['title']
    parents = [Obstemp, Progtemp, Multiple(Catalogue, maxnumber=20),
               Multiple(Targprog, maxnumber=20), Multiple(Survey, maxnumber=10),
               # Multiple('FibreTarget', maxnumber=999, notreal=True)
               ]
    idname = 'xml'  # this is CAT-NAME in the header not CATNAME, annoyingly no hyphens allowed


class FibreTarget(Hierarchy):
    """
    A fibretarget is the combination of fibre and surveytarget which is created after submission when
    the fibres are assigned to a target.
    This object describes where the fibre is placed and what its status is.
    It uniquely belongs to a surveytarget, fibre, and obspec.
    """
    factors = ['fibrera', 'fibredec', 'status', 'orientation', 'nretries', 'xposition', 'yposition',
               'targx', 'targy']
    parents = [Fibre, SurveyTarget]
    children = [OBSpec]
    identifier_builder = ['fibre', 'survey_target', 'obspec', 'xposition', 'yposition']


class OB(Hierarchy):
    """
    An OB is an "observing block" which is essentially a realisation of an OBSpec.
    Many OBs can share the same xml OBSpec which describes how to do the observations.
    """
    idname = 'id'  # This is globally unique by obid
    factors = ['mjd']
    parents = [OBSpec]


class Exposure(Hierarchy):
    """
    An exposure is one observation of one set of targets for a given configuration.
    WEAVE is structured such that an exposure is actually two sets of data, one from each arm.
    These are called runs.
    """
    idname = 'mjd'  # globally unique
    factors = ['exptime', 'seeingb', 'seeinge', 'windspb', 'windspe', 'telhumb', 'telhume',
               'winddirb', 'winddire', 'airmass',
               'teltempb', 'teltempe', 'skybrtel', 'observer', 'obstype']
    parents = [OB]


    @classmethod
    def from_header(cls, ob, header):
        factors = {f: header.get(f) for f in cls.factors}
        factors['mjd'] = np.round(float(header['MJD-OBS']), 6)
        factors['obstype'] = header['obstype'].lower()
        return cls(ob=ob, **factors)


class Spectrum(Hierarchy):
    is_template = True
    plural_name = 'spectra'


class Spectrum1D(Spectrum):
    is_template = True
    children = [OneOf(WavelengthHolder, one2one=True)]  # one2one because wvl only has 1 single & vice versa
    products = ['flux', 'ivar']
    plural_name = 'spectra1d'


class Spectrum2D(Spectrum):
    is_template = True
    plural_name = 'spectra2d'


class Run(Hierarchy):
    """
    A run is one observation of a set of targets for a given configuration in a specific arm (red or blue).
    A run belongs to an exposure, which always consists of one or two runs (per arm).
    """
    idname = 'id'
    parents = [Exposure, ArmConfig]
    children = [Optional('self', idname='adjunct', one2one=True)]


class RawSpectrum(Spectrum2D):
    """
    A 2D spectrum containing two counts arrays, this is not wavelength calibrated.
    """
    parents = [OneOf(Run, one2one=True)]
    products = ['counts1', 'counts2']
    children = [Optional('self', idname='adjunct', one2one=True)]
    identifier_builder = ['run']


class Single(Hierarchy):
    is_template = True


class Stacked(Hierarchy):
    is_template = True


class Stack(Stacked):
    is_template = True


class Superstack(Stacked):
    is_template = True


class Supertarget(Stacked):
    is_template = True


class MCMCMeasurement(BaseMeasurement):
    is_template = True
    factors = Measurement.factors + ['formal_error']


class Line(BaseMeasurement):
    is_template = True
    factors = ['aon']
    factors += Measurement.as_factors('flux', 'redshift', 'sigma', 'ebmv', 'amp')
    indexes = []


class SpectralIndex(BaseMeasurement):
    is_template = True


class RedshiftMeasurement(BaseMeasurement):
    is_template = True
    factors = Measurement.factors + ['warn']


class MeanFlux(Hierarchy):
    singular_name = 'mean_flux'
    plural_name = 'mean_fluxes'
    idname = 'band'
    factors = ['value']


def _predicate(o):
    if inspect.isclass(o):
        return issubclass(o, Hierarchy)
    return False
hierarchies = [i[-1] for i in inspect.getmembers(sys.modules[__name__], _predicate)]



import networkx as nx
import pytest

from weaveio.path_finding import HierarchyGraph
from weaveio.opr3.l1files import *
from weaveio.opr3.l2files import *
from weaveio.opr3.l1 import *
from weaveio.opr3.l2 import *

@pytest.fixture()
def graph():
    g = HierarchyGraph()
    g.initialise()
    return g


def make_path_test(graph, a, b):
    paths, _ = zip(*graph.find_paths(a, b, False))
    for path in paths:
        yield '-'.join(n.__name__ for n in path)


class _TestTemplate:
    a, b = None, None
    expected = None

    def test_path(self, graph):
        if self.expected is None:
            with pytest.raises(nx.NetworkXNoPath):
                make_path_test(graph, self.a, self.b)
        else:
            assert set(make_path_test(graph, self.a, self.b)) == self.expected

    def test_path_backwards(self, graph):
        if self.expected is None:
            with pytest.raises(nx.NetworkXNoPath):
                make_path_test(graph, self.b, self.a)
        else:
            assert set(make_path_test(graph, self.b, self.a)) == self.expected


class TestGandalfSurvey(_TestTemplate):
    a, b = Gandalf, Survey
    expected = {'Survey-Subprogramme-Catalogue-SurveyTarget-FibreTarget-L1Spectrum-Gandalf'}

class TestRVSpecfitSurvey(_TestTemplate):
    a, b = RVSpecfit, Survey
    expected = {'Survey-Subprogramme-Catalogue-SurveyTarget-FibreTarget-L1Spectrum-RVSpecfit'}

class TestRedrockSurvey(_TestTemplate):
    a, b = Redrock, Survey
    expected = {'Survey-Subprogramme-Catalogue-SurveyTarget-FibreTarget-L1Spectrum-Redrock'}

class TestFitSurvey(_TestTemplate):
    a, b = Fit, Survey
    expected = {'Survey-Subprogramme-Catalogue-SurveyTarget-FibreTarget-L1Spectrum-Fit'}

class TestFitIngestedSpectrum(_TestTemplate):
    a, b = Fit, IngestedSpectrum
    expected = {'IngestedSpectrum-ModelSpectrum-Fit'}

class TestFitUncombinedIngested(_TestTemplate):
    a, b = Fit, UncombinedIngestedSpectrum
    expected = {'UncombinedIngestedSpectrum-ModelSpectrum-Fit'}
    # TODO: this is actually ok for this database, but it feels wrong

class TestOBRun(_TestTemplate):
    a, b = OB, Run
    expected = {'OB-Exposure-Run'}

class TestL1SpectrumSurvey(_TestTemplate):
    a, b = L1Spectrum, Survey
    expected = {'Survey-Subprogramme-Catalogue-SurveyTarget-FibreTarget-L1Spectrum'}


class TestOBFit(_TestTemplate):
    a, b = OB, Fit
    expected = {
        'OB-Exposure-Run-RawSpectrum-L1SingleSpectrum-L1SupertargetSpectrum-Fit',
        'OB-Exposure-Run-RawSpectrum-L1SingleSpectrum-L1SuperstackSpectrum-Fit',
        'OB-L1StackSpectrum-Fit',
        'OB-Exposure-Run-RawSpectrum-L1SingleSpectrum-Fit',
    }

class TestOBArmConfig(_TestTemplate):
    a, b = ArmConfig, OB
    expected = {'ArmConfig-InstrumentConfiguration-Progtemp-OBSpec-OB'}


class TestArmconfigL1Single(_TestTemplate):
    a, b = ArmConfig, L1SingleSpectrum
    expected = {'ArmConfig-L1SingleSpectrum'}

class TestArmConfigL1Stack(_TestTemplate):
    a, b = ArmConfig, L1StackSpectrum
    expected = {'ArmConfig-L1StackSpectrum'}

class TestArmConfigL1Superstack(_TestTemplate):
    a, b = ArmConfig, L1SuperstackSpectrum
    expected = {'ArmConfig-L1SuperstackSpectrum'}

class TestArmConfigL1SuperTarget(_TestTemplate):
    a, b = ArmConfig, L1SupertargetSpectrum
    expected = {'ArmConfig-L1SupertargetSpectrum'}

class TestOBNoSS(_TestTemplate):
    a, b = OB, NoSS
    expected = {
        'OB-Exposure-Run-RawSpectrum-L1SingleSpectrum-NoSS',
        'OB-L1StackSpectrum-NoSS',
        'OB-Exposure-Run-RawSpectrum-L1SingleSpectrum-L1SuperstackSpectrum-NoSS',
        'OB-Exposure-Run-RawSpectrum-L1SingleSpectrum-L1SupertargetSpectrum-NoSS'
    }

class TestL1SpectrumTemplate(_TestTemplate):
    a, b = L1Spectrum, Template
    expected = {'L1Spectrum-Template'}

class TestRunWeaveTarget(_TestTemplate):
    a, b = Run, WeaveTarget
    expected = {'WeaveTarget-SurveyTarget-FibreTarget-OBSpec-OB-Exposure-Run'}

class TestWeaveTargetL1SingleSpectrum(_TestTemplate):
    a, b = L1SingleSpectrum, WeaveTarget
    expected = {'WeaveTarget-SurveyTarget-FibreTarget-L1SingleSpectrum'}

class TestWeaveTargetL1StackSpectrum(_TestTemplate):
    a, b = L1StackSpectrum, WeaveTarget
    expected = {'WeaveTarget-SurveyTarget-FibreTarget-L1StackSpectrum'}

class TestTemplate(_TestTemplate):
    a, b = Run, L1Spectrum
    expected = {
        'Run-RawSpectrum-L1SingleSpectrum',
        'Run-RawSpectrum-L1SingleSpectrum-L1StackSpectrum',
        'Run-RawSpectrum-L1SingleSpectrum-L1SuperstackSpectrum',
        'Run-RawSpectrum-L1SingleSpectrum-L1SupertargetSpectrum'
    }
#
# class TestTemplate2Template(_TestTemplate):
#     a, b = L1Spectrum, L1StackedSpectrum
#     expected = set()


from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from weaveio.readquery.parser import QueryGraph


class Statement:
    default_ids = ['inputs', '__class__']
    ids = []

    def __init__(self, input_variables, graph: 'QueryGraph', parameters=None):
        self.inputs = input_variables
        self.output_variables = []
        self._edge = None
        self.graph = graph
        self.parameters = parameters

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        return hash(self) == hash(o)

    def __hash__(self):
        ids = self.ids + self.default_ids
        obs = (getattr(self, i) for i in ids)
        obs = map(lambda x: tuple(x) if isinstance(x, list) else x, obs)
        return hash(tuple(map(hash, obs)))

    @property
    def edge(self):
        return self._edge

    @edge.setter
    def edge(self, value):
        self._edge = value
        self.graph.G.nodes[self._edge[1]]['variables'] = self.output_variables

    def make_variable(self, name):
        out = self.graph.get_variable_name(name)
        self.output_variables.append(out)
        return out

    def make_cypher(self, ordering: list) -> Optional[str]:
        raise NotImplementedError

    def __repr__(self):
        r = self.make_cypher(list(self.graph.G.nodes))
        if r is None:
            return 'Nothing'
        return f'{r}'


class StartingMatch(Statement):
    ids = ['to_node_type', 'unwound']

    def __init__(self, to_node_type, unwound, graph):
        super(StartingMatch, self).__init__([], graph)
        self.to_node_type = to_node_type
        self.unwound = unwound
        self.to_node = self.make_variable(to_node_type)

    def make_cypher(self, ordering: list) -> Optional[str]:
        return f"MATCH ({self.to_node}:{self.to_node_type})"


class Traversal(Statement):
    ids = ['to_node_type', 'string_paths', 'from_variable', 'unwound']

    def __init__(self, from_variable, to_node_type, paths, unwound, graph):
        super().__init__([from_variable], graph)
        self.from_variable = from_variable
        self.to_node_type = to_node_type
        self.paths = paths
        self.string_paths = '.'.join(paths)
        self.to_node = self.make_variable(to_node_type)
        self.using_edge = self.make_variable('edge')
        self.unwound = unwound  # this is only here to record dependency, the statement doesnt actually use it

    def make_cypher(self, ordering: list) -> str:
        if not self.string_paths:
            return f'OPTIONAL MATCH ({self.to_node}:{self.to_node_type})'
        paths = [path.format(in_var=self.from_variable, out_var=self.to_node, name=self.using_edge) for path in self.paths]
        statements = [f'OPTIONAL MATCH {path}' for path in paths]
        if len(statements) == 1:
            return statements[0]
        # use `union all` not `union` since there should be no duplicates as each end-node will be a different type
        statements = '\n\tUNION ALL\n'.join([f'\tWITH {self.from_variable}\n\t{s} RETURN {self.to_node}' for s in statements])
        return f"CALL {{\n {statements}\n}}"


class UnionTraversal(Statement):
    ids = ['to_node_type', 'path_names', 'path_values', 'from_variable', 'unwound']

    def __init__(self, from_variable, paths, to_node_type, unwound, graph):
        super().__init__([from_variable], graph)
        self.from_variable = from_variable
        self.to_node_type = to_node_type
        self.paths = paths
        self.path_names = list(paths.keys())
        self.path_values = list(paths.values())
        self.to_node = self.make_variable('node')
        self.using_edge = self.make_variable('edge')
        self.unwound = unwound

    def make_cypher(self, ordering: list) -> str:
        paths = [(n, path.format(name=self.using_edge)) for n, path in self.paths.items()]
        matches = 'UNION\n'.join([f'OPTIONAL MATCH ({self.from_variable}){path}({self.to_node}:{n})' for n, path in paths])
        return f"CALL {{WITH {self.from_variable}\n{matches}\nRETURN {self.to_node}\n}}"


class NullStatement(Statement):
    def __init__(self, input_variables, graph: 'QueryGraph'):
        super().__init__(input_variables, graph)
        self.output_variables = input_variables

    def make_cypher(self, ordering: list):
        return


class Operation(Statement):
    ids = ['op_string', 'op_name']

    def __init__(self, input_variable, dependency_variables, op_string, op_name, graph: 'QueryGraph', parameters: dict = None):
        super().__init__([input_variable]+dependency_variables, graph, parameters)
        self.op_string = op_string
        self.op_name = op_name
        self.op = f'({self.op_string.format(*self.inputs)})'
        self.output_variables.append(self.op)

    def make_cypher(self, ordering: list):
        return


class GetItem(Operation):
    ids = ['name']

    def __init__(self, input_variable, name, graph: 'QueryGraph'):
        super().__init__(input_variable, [], f'{{}}.`{name}`', f'.`{name}`', graph)
        self.name = name


class GetProduct(Operation):
    """
    Products are binary data that are not stored in the database.
    They are represented as a a relation between the parent node and an HDU of a fits file
    The relationship in neo4j will have the following properties: (product_name, column_name, index)
    Therefore, we return those properties after matching (node)<-[:PRODUCT {product_name:product_name}]-(hdu)
    """
    def __init__(self, input_variable, name, graph: 'QueryGraph'):
        super().__init__(input_variable, [], self.make_op_string(name), f'.{name}', graph)
        self.name = name

    def make_op_string(self, name):
        rel = f"({{}})<-[rel:product {{{{name: '{name}'}}}}]-(hdu: HDU)<--(file: File)"
        found = f"[{rel} | [file.path, hdu.extn, rel.index, rel.column_name, False]][0]"
        return found


class AssignToVariable(Statement):
    def __init__(self, input_variable, graph: 'QueryGraph'):
        super().__init__([input_variable], graph)
        self.input = input_variable
        self.output = self.make_variable('variable')

    def make_cypher(self, ordering: list) -> Optional[str]:
        return f"WITH *, {self.input} as {self.output}"


class Filter(Statement):
    def __init__(self, to_filter_variable, predicate_variable, graph: 'QueryGraph'):
        super().__init__([to_filter_variable, predicate_variable], graph)
        self.to_filter_variable = to_filter_variable
        self.predicate_variable = predicate_variable


class DirectFilter(Filter):
    def __init__(self, to_filter_variable, predicate_variable, graph: 'QueryGraph'):
        super().__init__(to_filter_variable, predicate_variable, graph)
        self.output = to_filter_variable
        self.output_variables = [to_filter_variable]

    def make_cypher(self, ordering: list) -> str:
        return f"WHERE {self.predicate_variable}"


class CopyAndFilter(Filter):
    def __init__(self, to_filter_variable, predicate_variable, graph: 'QueryGraph'):
        super().__init__(to_filter_variable, predicate_variable, graph)
        self.output = self.make_variable(to_filter_variable)

    def make_cypher(self, ordering: list) -> str:
        return f"WITH *, CASE WHEN {self.predicate_variable} THEN {self.to_filter_variable} ELSE null END as {self.output}"


class Slice(Statement):
    ids = ['slc']

    def __init__(self, input_variable, slc, graph: 'QueryGraph'):
        super().__init__([input_variable], graph)
        self.slc = slc

    def make_cypher(self, ordering: list) -> str:
        raise NotImplementedError


class Aggregate(Statement):
    ids = ['agg_func', 'name']

    def __init__(self, input, wrt_node, agg_func: str, name, graph: 'QueryGraph'):
        super().__init__([input, wrt_node], graph)
        self.agg_func = agg_func
        self.input = input
        self.wrt_node = wrt_node
        self.name = name
        self.output = self.make_variable(name)

    @property
    def conserve(self):
        above = self.graph.above_state(self.wrt_node)
        to_conserve = []
        for c in above:
            if not self.graph.node_holds_type(c, 'operation'):
                if self.graph.node_is_null_statement(c):
                    # this bit excludes fake aggregations where we have folded back along a single path (for consistency)
                    continue
                to_conserve.append(c)
        return to_conserve

    def make_cypher(self, ordering) -> str:
        conserve = {i for i in self.conserve if i in ordering}
        conserve = {self.graph.G.nodes[c]['variables'][0] for c in conserve if self.graph.G.nodes[c]['variables']}
        conserve = {c for c in conserve if c != self.output}
        variables = ', '.join(conserve) + ', ' if conserve else ''
        agg_string = self.agg_func.format(self.input)
        return f"WITH {variables}{agg_string} as {self.output}"


class Return(Statement):
    ids = ['index_variable', 'dropna']

    def __init__(self, column_variables, index_variable, dropna, graph: 'QueryGraph'):
        super().__init__(column_variables, graph)
        if index_variable is not None:
            self.inputs.append(index_variable)
        self.index_variable = index_variable
        self.column_variables = column_variables
        self.dropna = dropna

    def make_cypher(self, ordering: list) -> Optional[str]:
        names = [self.graph.get_variable_name('r') for _ in self.column_variables]  # to prevent overflow
        names.append(self.graph.get_variable_name('i'))
        cols = self.column_variables if self.index_variable is None else self.column_variables[:-1]
        cols = ', '.join([f"{c} as {n}" for c, n in zip(cols, names)])
        if self.dropna is not None:
            return f"WITH * WHERE {self.dropna} is not null RETURN {cols}"
        return f"RETURN {cols}"


class ApplyToList(Operation):
    def __init__(self, input_variable, dependency_variables, apply_function, filter_function,
                 graph: 'QueryGraph', put_null_in_empty=False, parameters=None):
        Statement.__init__(self, [input_variable, *dependency_variables], graph, parameters)
        self.x = graph.get_variable_name('x')
        self.op_name = 'list-apply'
        apply = apply_function.format(self.x, *self.inputs[1:])
        filt = filter_function.format(self.x, *self.inputs[1:])
        op_string = f"[{self.x} in {input_variable} WHERE {filt} | {apply}]"
        if put_null_in_empty:
            op_string = f"CASE WHEN SIZE({op_string}) = 0 THEN [null] ELSE {op_string} END"
        self.op_string = op_string
        self.op = op_string
        self.output_variables.append(self.op)



class Unwind(Statement):
    ids = ['parameter']

    def __init__(self, wrt, parameter, name, graph: 'QueryGraph', parameters: dict = None):
        super().__init__([wrt], graph, parameters)
        self.parameter = parameter
        self.output = self.make_variable(name)

    def make_cypher(self, ordering: list) -> Optional[str]:
        return f"UNWIND {self.parameter} as {self.output}"


from weaveio.readquery.utilities import remove_successive_duplicate_lines
import pytest
from string import printable
from hypothesis import given, strategies as st, example


@given(st.lists(st.text(printable, min_size=1, max_size=2)))
@example(['a', 'b', 'b'])
@example(['a', 'b', 'c'])
def test_remove_successive_duplicate_lines(x):
    removed = remove_successive_duplicate_lines(x)
    for a, b in zip(removed[:-1], removed[1:]):
        assert a != b
    pairs = list(zip(x[:-1], x[1:]))
    for a, b in pairs:
        if a != b:
            assert (a, b) in pairs



import inspect
import logging
from collections import OrderedDict, Counter
from copy import deepcopy
from functools import wraps, partial, reduce
from typing import Tuple, Dict, Type, Union, List, Optional as _Optional
from warnings import warn

import networkx as nx

from . import writequery
from .writequery import CypherQuery, Unwind, Collection, CypherVariable
from .context import ContextError
from .utilities import Varname, make_plural, int_or_none, camelcase2snakecase, snakecase2camelcase
from .writequery.base import CypherVariableItem, Alias


def _convert_types_to_node(x):
    if isinstance(x, dict):
        return {_convert_types_to_node(k): _convert_types_to_node(v) for k, v in x.items()}
    elif isinstance(x, (list, set, tuple)):
        return x.__class__([_convert_types_to_node(i) for i in x])
    elif isinstance(x, Graphable):
        return x.node
    else:
        return x

def hierarchy_query_decorator(function):
    @wraps(function)
    def inner(*args, **kwargs):
        args = _convert_types_to_node(args)
        kwargs = _convert_types_to_node(kwargs)
        return function(*args, **kwargs)
    return inner


unwind = hierarchy_query_decorator(writequery.unwind)
merge_node = hierarchy_query_decorator(writequery.merge_node)
match_node = hierarchy_query_decorator(writequery.match_node)
match_pattern_node = hierarchy_query_decorator(writequery.match_pattern_node)
match_id_node = hierarchy_query_decorator(writequery.match_id_node)
match_branch_node = hierarchy_query_decorator(writequery.match_branch_node)
collect = hierarchy_query_decorator(writequery.collect)
merge_relationship = hierarchy_query_decorator(writequery.merge_relationship)
set_version = hierarchy_query_decorator(writequery.set_version)
validate_number = hierarchy_query_decorator(writequery.validate_number)
validate_type = hierarchy_query_decorator(writequery.validate_type)


def chunker(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


FORBIDDEN_LABELS = []
FORBIDDEN_PROPERTY_NAMES = ['keys', 'values', 'dict', 'items']
FORBIDDEN_LABEL_PREFIXES = ['_']
FORBIDDEN_PROPERTY_PREFIXES = ['_']
FORBIDDEN_IDNAMES = ['idname']


class RuleBreakingException(Exception):
    pass


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


class Multiple:
    def __init__(self, node: Union[str, 'GraphableMeta'], minnumber=1, maxnumber=None, constrain=None, idname=None, one2one=False, ordered=False, notreal=False):
        """
        Used to define relationships in the schema .e.g.:
            class Car:
                parents = [Multiple(Wheel, 4, 4, one2one=True), Multiple(Driver, constrain=(House,)]
        This is read as:
            A car has exactly 4 wheels and a wheel has only one car
            A car has an unlimited number of drivers in the same house (but at least 1) and a driver has an unlimited number of cars (but at least 1)

        :param node: A Hierarchy object to define a relationship to (e.g. creating a rel of Car-->Wheel
        :param minnumber: The minimum number that can exists in the database, can be 0
        :param maxnumber: The rough advised number of relationships that can exist
        :param constrain: The Hierarchies which all node instances must have in their tree of dependencies
        :param idname: The name of the relationship which can be used when reading
        :param one2one: Is the relationship reciprocated? If yes, then a relationship of parent-[singular]->node  is also created
                        e.g. After the Wheel-->Car rel is created, Car-[singular]->Wheel is also created.
                            In this case a car has multiple wheels, but a wheel has only 1 car.
        :param ordered: The order of the instances makes a different unique identification
        :param notreal: This relationship is never instantiated. Use this to express logic about superclasses and subclasses
        """
        if not isinstance(node, (GraphableMeta, str)):
            raise TypeError(f"{node} is of type {type(node)}. Must be a Hierarchy class or a name of a yet to be built Hierarchy class")
        self.node = node
        self.minnumber = int_or_none(minnumber) or 0
        if maxnumber is None:
            warn(f"maxnumber is None for {node}", RuntimeWarning)
        self.maxnumber = int_or_none(maxnumber)
        self.constrain = tuple() if constrain is None else (constrain, ) if not isinstance(constrain, (list, tuple)) else tuple(constrain)
        self.relation_idname = idname
        self.one2one = one2one
        self._isself = self.node == 'self'
        self.notreal = notreal
        self.ordered = bool(ordered)
        if inspect.isclass(self.node):
            if issubclass(self.node, Hierarchy):
                self.instantate_node()

    def to_reldict(self):
        return {'relation_idname': self.relation_idname, 'maxnumber': self.maxnumber, 'minnumber': self.minnumber,
                'constrain': self.constrain, 'one2one': self.one2one, 'notreal': self.notreal, 'ordered': self.ordered}

    @property
    def is_optional(self):
        return self.minnumber == 0

    @property
    def name(self):
        if self.relation_idname is not None:
            return self.relation_idname
        if self.maxnumber == 1:
            return self.singular_name
        return self.plural_name

    def instantate_node(self, include_hierarchies=None):
        if not inspect.isclass(self.node):
            if isinstance(self.node, str):
                hierarchies = {i.__name__: i for i in all_subclasses(Hierarchy)}
                if include_hierarchies is not None:
                    for h in include_hierarchies:
                        hierarchies[h.__name__] = h  # this overrides the default
                try:
                    self.node = hierarchies[self.node]
                except KeyError:
                    # require hierarchy doesnt exist yet
                    Hierarchy._waiting.append(self)
                    return
        self.singular_name = self.node.singular_name
        self.plural_name = self.node.plural_name

        try:
            self.factors =  self.node.factors
        except AttributeError:
            self.factors = []
        try:
            self.parents = self.node.parents
        except AttributeError:
            self.parents = []
        while Hierarchy._waiting:
            h = Hierarchy._waiting.pop()
            h.instantate_node(include_hierarchies)

    def __repr__(self):
        if self.ordered:
            o = '[Ordered]'
        else:
            o = '[Unordered]'
        return f"<Multiple({self.node} [{self.minnumber} - {self.maxnumber}] id={self.relation_idname} {o})>"

    def __hash__(self):
        if isinstance(self.node, str):
            hsh = hash(self.node)
        else:
            hsh = hash(self.node.__name__)
        return hash(self.__class__) ^ hash(self.minnumber) ^ hash(self.maxnumber) ^ hash(self.ordered) ^ \
        reduce(lambda x, y: x ^ y, map(hash, self.constrain), 0) ^ hash(self.relation_idname) ^ hsh

    def __eq__(self, other):
        return hash(self) == hash(other)

    @classmethod
    def from_names(cls, hierarchy: Type['Hierarchy'], *singles: str, **multiples: Union[int, Tuple[_Optional[int], _Optional[int]]]) -> List['Multiple']:
        single_list = [OneOf(hierarchy, idname=name) for name in singles]
        multiple_list = [Multiple(hierarchy, i, i) if isinstance(i, int) else Multiple(cls, *i) for k, i in multiples.items()]
        return single_list + multiple_list

    @classmethod
    def from_any(cls, anything: Union[str, Type['Hierarchy'], 'Multiple']) -> 'Multiple':
        if isinstance(anything, (str, GraphableMeta)):
            return OneOf(anything)
        elif isinstance(anything, Multiple):
            return anything
        raise TypeError(f"{anything} is not a subclass of Hierarchy or an instance of Multiple")

class OneOf(Multiple):
    def __init__(self, node, constrain=None, idname=None, one2one=False):
        super().__init__(node, 1, 1, constrain, idname, one2one)

    def __repr__(self):
        return f"<OneOf({self.node})>"


class Optional(Multiple):
    def __init__(self, node, constrain=None, idname=None, one2one=False):
        super(Optional, self).__init__(node, 0, 1, constrain, idname, one2one)

    def __repr__(self):
        return f"<Optional({self.node})>"


class GraphableMeta(type):
    def __new__(meta, name: str, bases, _dct):
        dct = {'is_template': False}
        dct.update(_dct)
        dct['aliases'] = dct.get('aliases', [])
        dct['aliases'] += [a for base in bases for a in base.aliases]
        dct['singular_name'] = dct.get('singular_name', None) or camelcase2snakecase(name)
        dct['plural_name'] = dct.get('plural_name', None) or make_plural(dct['singular_name'])
        if dct['plural_name'] != dct['plural_name'].lower():
            raise RuleBreakingException(f"plural_name must be lowercase")
        if dct['singular_name'] != dct['singular_name'].lower():
            raise RuleBreakingException(f"singular_name must be lowercase")
        if dct['plural_name'] == dct['singular_name']:
            raise RuleBreakingException(f"plural_name must not be the same as singular_name")
        dct['name'] = dct['singular_name']
        idname = dct.get('idname', None)
        if idname in FORBIDDEN_IDNAMES:
            raise RuleBreakingException(f"You may not name an id as one of {FORBIDDEN_IDNAMES}")
        if not (isinstance(idname, str) or idname is None):
            raise RuleBreakingException(f"{name}.idname ({idname}) must be a string or None")
        if name[0] != name.capitalize()[0] or '_' in name:
            raise RuleBreakingException(f"{name} must have `CamelCaseName` style name")
        for factor in dct.get('factors', []) + ['idname'] + [dct['singular_name'], dct['plural_name']]:
            # if factor != factor.lower():
            #     raise RuleBreakingException(f"{name}.{factor} must have `lower_snake_case` style name")
            if factor in FORBIDDEN_PROPERTY_NAMES:
                raise RuleBreakingException(f"The name {factor} is not allowed for class {name}")
            if any(factor.startswith(p) for p in FORBIDDEN_PROPERTY_PREFIXES):
                raise RuleBreakingException(f"The name {factor} may not start with any of {FORBIDDEN_PROPERTY_PREFIXES} for {name}")
            if not factor.islower():
                raise RuleBreakingException(f"The name {factor} must be lowercase")
        # remove duplicates from the list dct['parents'] whilst maintaining its order
        if 'parents' in dct:
            dct['parents'] = list(OrderedDict.fromkeys(dct['parents']))
        if 'children' in dct:
            dct['children'] = list(OrderedDict.fromkeys(dct['children']))
        if 'factors' in dct:
            dct['factors'] = list(OrderedDict.fromkeys(dct['factors']))
        if 'produces' in dct:
            dct['produces'] = list(OrderedDict.fromkeys(dct['produces']))
        r = super(GraphableMeta, meta).__new__(meta, name, bases, dct)
        return r

    def __init__(cls, name, bases, dct):
        invalid_bases = [b for b in bases if not b.is_template]
        if any(invalid_bases):
            raise RuleBreakingException(f"{name} cannot subclass {invalid_bases} since they are marked with `is_template=False`")
        if cls.idname is not None and cls.identifier_builder is not None:
            raise RuleBreakingException(f"You cannot define a separate idname and an identifier_builder at the same time for {name}")
        parentnames = {}
        cls.children = deepcopy(cls.children)  # sever link so that changes here dont affect base classes
        cls.parents = deepcopy(cls.parents)
        for i, c in enumerate(cls.children):
            if isinstance(c, Multiple):
                if c._isself:
                    c.node = cls
                c.instantate_node()
                for n in c.constrain:
                    if n not in cls.children:
                        cls.children.append(n)
                parentnames[c.name] = (c.minnumber, c.maxnumber)
            else:
                parentnames[c.singular_name] = (1, 1)
        for i, p in enumerate(cls.parents):
            if isinstance(p, Multiple):
                if p._isself:
                    p.node = cls
                p.instantate_node()
                for n in p.constrain:
                    if n not in cls.parents:
                        cls.parents.append(n)
                parentnames[p.name] = (p.minnumber, p.maxnumber)
            else:
                parentnames[p.singular_name] = (1, 1)
        if cls.identifier_builder is not None:
            for p in cls.identifier_builder:
                if isinstance(p, type):
                    if issubclass(p, Hierarchy):
                        p = p.singular_name
                if p in parentnames:
                    mn, mx = parentnames[p]
                elif p in cls.factors:
                    pass
                else:
                    raise RuleBreakingException(f"Unknown identifier source {p} for {name}. "
                                                f"Available are: {list(parentnames.keys())+cls.factors}")
        version_parents = []
        version_factors = []
        for p in cls.version_on:
            if p in [pp.singular_name if isinstance(pp, type) else pp.name for pp in cls.parents+cls.children]:
                version_parents.append(p)
            elif p in cls.factors:
                version_factors.append(p)
            else:
                raise RuleBreakingException(f"Unknown {p} to version on for {name}. Must refer to a parent or factor.")
        if len(version_factors) > 1 and len(version_parents) == 0:
            raise RuleBreakingException(f"Cannot build a version relative to nothing. You must version on at least one parent.")
        if not cls.is_template:
            if not (len(cls.indexes) or cls.idname or
                    (cls.identifier_builder is not None and len(cls.identifier_builder) > 0)):
                raise RuleBreakingException(f"{name} must define an indexes, idname, or identifier_builder")
        for p in cls.indexes:
            if p is not None:
                if p not in cls.parents and p not in cls.factors:
                    raise RuleBreakingException(f"index {p} of {name} must be a factor or parent of {name}")
        if cls.concatenation_constants is not None:
            if len(cls.concatenation_constants):
                cls.factors = cls.factors + cls.concatenation_constants + ['concatenation_constants']
        clses = [i.__name__ for i in inspect.getmro(cls)]
        clses = clses[:clses.index('Graphable')]
        cls.neotypes = clses
        cls.products_and_factors = cls.factors + cls.products
        if cls.idname is not None:
            cls.products_and_factors.append(cls.idname)
        cls.relative_names = {}  # reset, no inheritability
        for relative in cls.children+cls.parents:
            if isinstance(relative, Multiple):
                if relative.relation_idname is not None:
                    cls.relative_names[relative.relation_idname] = relative
        cls.indexable = cls.idname is not None or cls.identifier_builder is not None
        super().__init__(name, bases, dct)


class Graphable(metaclass=GraphableMeta):
    idname = None
    identifier = None
    indexer = None
    indexable = False
    type_graph_attrs = {}
    plural_name = None
    singular_name = None
    parents = []
    children = []
    uses_tables = False
    factors = []
    data = None
    query = None
    is_template = True
    products = []
    indexes = []
    identifier_builder = None
    version_on = []
    produces = []
    concatenation_constants = []
    belongs_to = []
    products_and_factors = []
    relative_names = {}

    @property
    def node(self):
        return self._node

    @node.setter
    def node(self, value):
        assert isinstance(value, CypherVariable)
        self._node = value

    @classmethod
    def requirement_names(cls):
        l = []
        for p in cls.parents:
            if isinstance(p, type):
                if issubclass(p, Graphable):
                    l.append(p.singular_name)
            else:
                if isinstance(p, Multiple):
                    if p.maxnumber == 1:
                        l.append(p.singular_name)
                    else:
                        l.append(p.plural_name)
                else:
                    raise RuleBreakingException(f"The parent list of a Hierarchy must contain "
                                                f"only other Hierarchies or Multiple(Hierarchy)")

        return l

    def add_parent_data(self, data):
        self.data = data

    def add_parent_query(self, query):
        self.query = query

    def __getattr__(self, item):
        if self.query is not None:
            warn('Lazily loading a hierarchy attribute can be costly. Consider using a more flexible query.')
            attribute = getattr(self.query, item)()
            setattr(self, item, attribute)
            return attribute
        raise AttributeError(f"Query not added to {self}, cannot search for {self}.{item}")

    @property
    def neoproperties(self):
        identifier_builder = [] if self.identifier_builder is None else self.identifier_builder
        d = {}
        for f in self.factors:
            if f not in identifier_builder and f != self.idname:
                value = getattr(self, f.lower())
                if value is not None:
                    d[f.lower()] = value
        return d

    @property
    def neoidentproperties(self):
        identifier_builder = [] if self.identifier_builder is None else self.identifier_builder
        d = {}
        if self.identifier is None and self.idname is not None:
            raise ValueError(f"{self} must have an identifier")
        if self.idname is None and self.identifier is not None:
            raise ValueError(f"{self} must have an idname to be given an identifier")
        elif self.idname is not None:
            d[self.idname] = self.identifier
            d['id'] = self.identifier
        for f in self.factors:
            if f in identifier_builder:
                value = getattr(self, f.lower())
                if value is not None:
                    d[f.lower()] = value
        return d


    def __init__(self, predecessors, successors=None, do_not_create=False):
        if self.is_template and not do_not_create:
            raise TypeError(f"Template Hierarchies cannot be instantiated directly. "
                            f"If you need to instantiate {self.__class__.__name__}, set {self.__class__.__name__}.is_template=False.")
        if successors is None:
            successors = {}
        if predecessors is None:
            predecessors = {}
        self.predecessors = predecessors
        self.successors = successors
        self.data = None
        if do_not_create:
            return
        try:
            query = CypherQuery.get_context()  # type: CypherQuery
            collision_manager = query.collision_manager
        except ContextError:
            return
        merge_strategy = self.__class__.merge_strategy()
        version_parents = []
        if  merge_strategy == 'NODE FIRST':
            self.node = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
                                           collision_manager=collision_manager)
            for k, (parent_list, relation) in predecessors.items():
                type = 'is_required_by'
                if isinstance(parent_list, Collection):
                    with unwind(parent_list, enumerated=True) as (parent, i):
                        props = {'relation_id': k}
                        if relation.ordered:
                            props['order'] = i
                        merge_relationship(parent, self.node, type, {}, props, collision_manager=collision_manager)
                    parent_list = collect(parent)
                    if k in self.version_on:
                        raise RuleBreakingException(f"Cannot version on a collection of nodes")
                else:
                    for i, parent in enumerate(parent_list):
                        props = {'relation_id': k}
                        if relation.ordered:
                            props['order'] = i
                        merge_relationship(parent, self.node, type, {}, props, collision_manager=collision_manager)
                        if k in self.version_on:
                            version_parents.append(parent)
            # now the children
            for k, (child_list, relation) in successors.items():
                type = 'is_required_by'
                if isinstance(child_list, Collection):
                    with unwind(child_list, enumerated=True) as (child, i):
                        props = {'relation_id': k}
                        if relation.ordered:
                            props['order'] = i
                        merge_relationship(self.node, child, type,
                                           {}, props,
                                           collision_manager=collision_manager)
                    collect(child)
                else:
                    for i, child in enumerate(child_list):
                        props = {'relation_id': k}
                        if relation.ordered:
                            props['order'] = i
                        merge_relationship(self.node, child, type,
                                           {}, props,
                                           collision_manager=collision_manager)
        elif merge_strategy == 'NODE+RELATIONSHIP':
            parentnames = {p.name: p for p in self.parents}
            childnames = {p.name: p for p in self.children}
            parents = []
            children = []
            other_parents = []
            other_children = []
            for k, (parent_list, relation) in self.predecessors.items():
                if k in self.identifier_builder and k in parentnames:
                    if isinstance(parent_list, Collection):
                        raise TypeError(f"Cannot merge NODE+RELATIONSHIP for collections")
                    if relation.ordered:
                        parents += [({'order': i, 'relation_id': k}, p) for i, p in enumerate(parent_list)]
                    else:
                        parents += [({'relation_id': k}, p) for i, p in enumerate(parent_list)]
                elif isinstance(parent_list, Collection):
                    if relation.ordered:
                        other_parents.append(({'order': None, 'relation_id': k}, parent_list))
                    else:
                        other_parents.append(({'relation_id': k}, parent_list))
                else:
                    if relation.ordered:
                        other_parents += [({'order': i, 'relation_id': k}, p) for i, p in enumerate(parent_list)]
                    else:
                        other_parents += [({'relation_id': k}, p) for i, p in enumerate(parent_list)]
                if k in self.version_on:
                    raise NotImplementedError(f"Versioning is not yet implemented: {k} cannot be versioned yet")
            for k, (child_list, relation) in self.successors.items():
                if k in self.identifier_builder and k in childnames:
                    if isinstance(child_list, Collection):
                        raise TypeError(f"Cannot merge NODE+RELATIONSHIP for collections")
                    if relation.ordered:
                        children += [({'order': i, 'relation_id': k}, c) for i, c in enumerate(child_list)]
                    else:
                        children += [({'relation_id': k}, c) for i, c in enumerate(child_list)]
                elif isinstance(child_list, Collection):
                    if relation.ordered:
                        other_children.append(({'order': None, 'relation_id': k}, child_list))
                    else:
                        other_children.append(({'relation_id': k}, child_list))
                else:
                    if relation.ordered:
                        other_children += [({'order': i, 'relation_id': k}, c) for i, c in enumerate(child_list)]
                    else:
                        other_children += [({'relation_id': k}, c) for i, c in enumerate(child_list)]
                if k in self.version_on:
                    raise NotImplementedError(f"Versioning is not yet implemented: {k} cannot be versioned yet")
            reltype = 'is_required_by'
            rels = {p: (reltype, True, d, {}) for d, p in parents}
            rels.update({c: (reltype, False, d, {}) for d, c in children})
            anti_id_rels = []
            # make sure that optional relations present in an ID are part of the search
            # i.e. the absence of an optional relation in the db is itself part of the unique key
            # anti_id_rels are added for missing optional relations
            for k in self.identifier_builder:
                if k not in self.predecessors and k not in self.successors:
                    p = parentnames.get(k, None)
                    c = childnames.get(k, None)
                    if getattr(p, 'is_optional', False):
                        anti_id_rels.append((p.node.__name__, reltype, True, {'relation_id': p.relation_idname or k}))
                    if getattr(c, 'is_optional', False):
                        anti_id_rels.append((c.node.__name__, reltype, False, {'relation_id': p.relation_idname or k}))
            self.node = merge_node(self.neotypes, self.neoidentproperties, self.neoproperties,
                                           id_rels=rels, anti_id_rels=anti_id_rels, collision_manager=collision_manager)
            for d, others in other_parents:
                if others is not None:
                    if isinstance(others, Collection):
                        with unwind(others, enumerated=True) as (other, i):
                            if 'order' in d:
                                d['order'] = i
                            merge_relationship(other, self.node, reltype, d, {}, collision_manager=collision_manager)
                        collect(other)
                    else:
                        merge_relationship(others, self.node, reltype, d, {}, collision_manager=collision_manager)
            for d, others in other_children:
                if others is not None:
                    if isinstance(others, Collection):
                        with unwind(others, enumerated=True) as (other, i):
                            if 'order' in d:
                                d['order'] = i
                            merge_relationship(self.node, other, reltype, d, {}, collision_manager=collision_manager)
                        collect(other)
                    else:
                        merge_relationship(self.node, others, reltype, d, {}, collision_manager=collision_manager)
        else:
            ValueError(f"Merge strategy not known: {merge_strategy}")
        if len(version_parents):
            raise NotImplementedError(f"Versioning is not yet implemented: {version_parents} cannot be versioned yet")


    @classmethod
    def has_factor_identity(cls):
        if cls.identifier_builder is None:
            return False
        if len(cls.identifier_builder) == 0:
            return False
        rels = [x.name if isinstance(x, Multiple) else OneOf(x).name for x in cls.parents+cls.children]
        if not any(r in cls.factors or r in cls.identifier_builder for r in rels) and rels:
            raise ValueError(f"{cls} defines an identifier_builder with name(s) which are neither factors, parents, or children")
        return not any(r in cls.identifier_builder for r in rels)

    @classmethod
    def has_rel_identity(cls):
        if cls.identifier_builder is None:
            return False
        if len(cls.identifier_builder) == 0:
            return False
        return any(n.name in cls.identifier_builder for n in cls.parents + cls.children)

    @classmethod
    def make_schema(cls) -> List[str]:
        name = cls.__name__
        indexes = []
        nonunique = False
        if cls.is_template:
            return []
        elif cls.idname is not None:
            prop = cls.idname
            indexes.append(f'CREATE CONSTRAINT {name}_id ON (n:{name}) ASSERT (n.{prop}) IS NODE KEY')
        elif cls.identifier_builder:
            if cls.has_factor_identity():  # only of factors
                key = ', '.join([f'n.{f}' for f in cls.identifier_builder])
                indexes.append(f'CREATE CONSTRAINT {name}_id ON (n:{name}) ASSERT ({key}) IS NODE KEY')
            elif cls.has_rel_identity():  # based on rels from parents/children
                # create 1 index on id factors and 1 index per factor as well
                key = ', '.join([f'n.{f}' for f in cls.identifier_builder if f in cls.factors])
                if key:  # joint index
                    indexes.append(f'CREATE INDEX {name}_rel FOR (n:{name}) ON ({key})')
                # separate indexes
                indexes += [f'CREATE INDEX {name}_{f} FOR (n:{name}) ON (n.{f})' for f in cls.identifier_builder if f in cls.factors]
        else:
            nonunique = True
        if cls.indexes:
            id = cls.identifier_builder or []
            indexes += [f'CREATE INDEX {name}_{i} FOR (n:{name}) ON (n.{i})' for i in cls.indexes if i not in id]
        if not indexes and nonunique:
            raise RuleBreakingException(f"{name} must define an idname, identifier_builder, or indexes, "
                                        f"unless it is marked as template class for something else (`is_template=True`)")
        return indexes

    @classmethod
    def merge_strategy(cls):
        if cls.idname is not None:
            return 'NODE FIRST'
        elif cls.identifier_builder:
            if cls.has_factor_identity():
                return 'NODE FIRST'
            elif cls.has_rel_identity():
                return 'NODE+RELATIONSHIP'
        return 'NODE FIRST'

    def attach_product(self, product_name, hdu, index=None, column_name=None):
        """attaches products to a hierarchy with relations like: <-[:PRODUCT {index: rowindex, name: 'flux'}]-"""
        if product_name not in self.products:
            raise TypeError(f"{product_name} is not a product of {self.__class__.__name__}")
        collision_manager = CypherQuery.get_context().collision_manager
        props = {'name': product_name}
        if index is not None:
            props['index'] = index
        if column_name is not None:
            props['column_name'] = column_name
        merge_relationship(hdu, self, 'product', props, {}, collision_manager=collision_manager)

    @classmethod
    def without_creation(cls, **kwargs):
        return cls(do_not_create=True, **kwargs)

    @classmethod
    def find(cls, anonymous_children=None, anonymous_parents=None,
             exclude=None,
             **kwargs):
        parent_names = [i.name if isinstance(i, Multiple) else i.singular_name for i in cls.parents]
        parents = [] if anonymous_parents is None else anonymous_parents
        anonymous_children = [] if anonymous_children is None else anonymous_children
        factors = {}
        for k, v in kwargs.items():
            if k in cls.factors:
                factors[k] = v
            elif k in parent_names:
                if not isinstance(v, list):
                    v = [v]
                for vi in v:
                    parents.append(vi)
            elif k == cls.idname:
                factors[k] = v
            else:
                raise ValueError(f"Unknown name {k} for {cls}")
        node = match_pattern_node(labels=cls.neotypes, properties=factors,
                                  parents=parents, children=anonymous_children, exclude=exclude)
        obj = cls.without_creation(**kwargs)
        obj.node = node
        return obj


    def __repr__(self):
        i = ''
        if self.idname is not None:
            i = f'{self.identifier}'
        return f"<{self.__class__.__name__}({self.idname}={i})>"


class Hierarchy(Graphable):
    parents = []
    factors = []
    _waiting = []
    _hierarchies = {}
    is_template = True


    @classmethod
    def from_cypher_variable(cls, variable):
        thing = cls(do_not_create=True)
        thing.node = variable
        return thing

    @classmethod
    def from_neo4j_id(cls, id):
        node = match_id_node(labels=cls.neotypes, id=id)
        return cls.from_cypher_variable(node)

    @classmethod
    def as_factors(cls, *names, prefix=''):
        if len(names) == 1 and isinstance(names[0], list):
            names = prefix+names[0]
        if cls.parents+cls.children:
            raise TypeError(f"Cannot use {cls} as factors {names} since it has defined parents and children")
        return [f"{prefix}{name}_{factor}".lower() if factor != 'value' else f"{prefix}{name}".lower() for name in names for factor in cls.factors]

    @classmethod
    def from_name(cls, name):
        singular_name = f"{name}_{cls.singular_name}"
        plural_name = f"{name}_{cls.plural_name}"
        name = snakecase2camelcase(name)
        name = f"{name}{cls.__name__}"
        try:
            return cls._hierarchies[name]
        except KeyError:
            cls._hierarchies[name] = type(name, (cls,), {'singular_name': singular_name, 'plural_name': plural_name})
            return cls._hierarchies[name]

    @classmethod
    def from_names(cls, *names):
        if len(names) == 1 and isinstance(names[0], list):
            names = names[0]
        return [cls.from_name(name) for name in names]

    def make_specification(self) -> Tuple[Dict[str, Type[Graphable]], Dict[str, str], Dict[str, Type[Graphable]], Dict[str, Type[Graphable]]]:
        """
        Make a dictionary of {name: HierarchyClass} and a similar dictionary of factors
        """
        # ordered here since we need to use the first parent as a match point in merge_dependent_node
        parents = OrderedDict()
        children = OrderedDict()
        for inl, outl, mirrorl in ([self.parents, parents, children], [self.children, children, parents]):
            for p in inl:
                name = getattr(p, 'relation_idname', None) or getattr(p, 'name', None) or p.singular_name
                outl[name] = p
                if isinstance(p, Multiple):
                    if p.one2one:
                        mirrorl[name] = p
        factors = {f.lower(): f for f in self.factors}
        specification = parents.copy()
        specification.update(factors)
        specification.update(children)
        return specification, factors, parents, children

    @classmethod
    def instantate_nodes(cls, hierarchies=None):
        for i in cls.parents + cls.factors + cls.children:
            if isinstance(i, Multiple):
                if isinstance(i.node, str):
                    i.instantate_node(hierarchies)

    def __init__(self, do_not_create=False, tables=None, tables_replace: Dict = None,
                 **kwargs):
        if tables_replace is None:
            tables_replace = {}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        self.instantate_nodes()
        self.uses_tables = False
        if tables is None:
            for value in kwargs.values():
                if isinstance(value, Unwind):
                    self.uses_tables = True
                elif isinstance(value, Hierarchy):
                    self.uses_tables = value.uses_tables
        else:
            self.uses_tables = True
        self.identifier = kwargs.pop(self.idname, None)
        self.specification, factors, parents, children = self.make_specification()
        # add any data held in a neo4j unwind table
        if tables is not None:
            for k, v in factors.items():
                if k not in kwargs:
                        kwargs[k] = tables.get(tables_replace.get(k, k), alias=False)
        # Make predecessors a dict of {name: [instances of required Factor/Hierarchy]}
        predecessors = {}
        successors = {}
        for name, nodetype in self.specification.items():
            if isinstance(nodetype, Multiple):
                if nodetype.notreal and name in kwargs:
                    raise ValueError(f"{name} is not accepted as an argument for {self} since `notreal` is set for it in {self.__class__.__name__} definition")
                if (nodetype.minnumber == 0 and name not in kwargs) or nodetype.notreal:
                    continue
            if do_not_create:
                value = kwargs.pop(name, None)
            else:
                value = kwargs.pop(name)
            setattr(self, name, value)
            if isinstance(nodetype, Multiple) and not isinstance(nodetype, (OneOf, Optional)):
                if nodetype.maxnumber != 1:
                    if not isinstance(value, (tuple, list)):
                        if isinstance(value, Graphable):
                            if not getattr(value, 'uses_tables', False):
                                raise TypeError(f"{name} expects multiple elements")
            else:
                value = [value]
            try:
                l = len(value)
            except TypeError:
                l = 1
            nodetype = Multiple.from_any(nodetype)
            if not do_not_create:
                if not isinstance(value, Collection) and not isinstance(getattr(value, 'node', None), Collection):
                    if not (nodetype.minnumber <= l <= nodetype.maxnumber):
                        raise ValueError(f"{self.__class__.__name__} requires {nodetype.minnumber} <= count({name}) <= {nodetype.maxnumber}. "
                                         f"{l} received.")
                    if isinstance(value, Hierarchy):
                        invalid_types = {v for v in value if isinstance(v, nodetype.node)}
                        if invalid_types:
                            raise TypeError(f"{self.__class__.__name__} can only take {nodetype.node.__name__} not {invalid_types}")
                else:
                    validate_number(value, nodetype.minnumber, nodetype.maxnumber, self.__class__.__name__, nodetype.node.__name__)
                    validate_type(value, nodetype.node.__name__)
            if name not in factors:
                if name in children:
                    successors[name] = value, nodetype
                if name in parents:
                    predecessors[name] = value, nodetype
        if len(kwargs) and not do_not_create:
            raise KeyError(f"{kwargs.keys()} are not relevant to {self.__class__}")
        self.predecessors = predecessors
        self.successors = successors
        if self.identifier_builder is not None:
            if self.identifier is not None:
                raise RuleBreakingException(f"{self} must not take an identifier if it has an identifier_builder")
        if self.idname is not None:
            if not do_not_create and self.identifier is None:
                raise ValueError(f"Cannot assign an id of None to {self}")
            setattr(self, self.idname, self.identifier)
        super(Hierarchy, self).__init__(predecessors, successors, do_not_create)

    def __getitem__(self, item):
        assert isinstance(item, str)
        getitem = CypherVariableItem(self.node, item)
        query = CypherQuery.get_context()
        if isinstance(item, int):
            item = f'{self.namehint}_index{item}'
        alias_statement = Alias(getitem, str(item))
        query.add_statement(alias_statement)
        return alias_statement.out

    def attach_optionals(self, **optionals):
        try:
            query = CypherQuery.get_context()  # type: CypherQuery
            collision_manager = query.collision_manager
        except ContextError:
            return
        reltype = 'is_required_by'
        parents = {p.name: p.one2one for p in self.parents if isinstance(p, Multiple) and p.is_optional}
        children = {c.name: c.one2one for c in self.children if isinstance(c, Multiple) and c.is_optional}
        for key, item in optionals.items():
            if isinstance(item, Collection):
                raise NotImplementedError(f"Cannot attach optional collections yet")
            if key in parents:
                parent, child = item.node, self.node
                one2one = parents[key]
            elif key in children:
                parent, child = self.node, item.node
                one2one = children[key]
            else:
                raise KeyError(f"{key} is not a parent or child of {self}")
            merge_relationship(parent, child, reltype, {'order': 0, 'relation_id': key}, {}, collision_manager=collision_manager)
            if one2one:
                merge_relationship(child, parent, reltype, {'order': 0, 'relation_id': key}, {}, collision_manager=collision_manager)

def find_branch(*nodes_or_types):
    """
    Given a mix of variables and types along an undirected path (the input order), instantate those types from the graph
    >>> _, l1spectrum, fibretargetr, _ = find_branch(l1file, L1Spectrum, FibreTarget, fibre)
    This will find the branch that looks like (l1file)<--(:L1Spectrum)<--(:FibreTarget)<--(fibre)
    and return the nodes
    """
    nodes = match_branch_node(*[i.__name__ if isinstance(i, type) else i for i in nodes_or_types])
    instances = []
    for node, node_or_type in zip(nodes, nodes_or_types):
        if isinstance(node_or_type, GraphableMeta):
            instance = node_or_type.from_cypher_variable(node)
        elif isinstance(node_or_type, Hierarchy):
            instance = node_or_type.__class__.from_cypher_variable(node)
        else:
            raise TypeError(f"Unknown type {type(node_or_type)}")
        instances.append(instance)
    return instances



from typing import List, TYPE_CHECKING

import textdistance
import networkx as nx

if TYPE_CHECKING:
    from .base import BaseQuery
    from .objects import ObjectQuery

__all__ = ['attributes', 'objects', 'explain', 'find']


def attributes(query: 'BaseQuery', plural=False, directly_owned_only=False) -> List[str]:
    from ..data import Data
    from .objects import ObjectQuery
    if isinstance(query, Data):
        return list(query.plural_factors.keys())
    if not isinstance(query, ObjectQuery):
        raise TypeError(f"{query} is not an ObjectQuery or Data")
    if directly_owned_only:
        return sorted([i.lower() for i in query._data.class_hierarchies[query._obj].products_and_factors])
    o = query._data.class_hierarchies[query._obj]
    def _test(n):
        if n in G:
            return nx.has_path(G, o, n)
        return False
    G = query._data.hierarchy_graph.singular
    reachable = set(G.subgraph_view(_test).nodes) - {o}
    attrs = {f'{n.singular_name}.{i.lower()}' for n in reachable for i in n.products_and_factors}
    if plural:
        G = query._data.hierarchy_graph.plural
        reachable = set(G.subgraph_view(_test).nodes) - {o}
        attrs |= {f'{n.plural_name}.{i.lower()}' for n in reachable for i in n.products_and_factors}
    return sorted([i.lower() for i in o.products_and_factors]) + sorted(attrs)

def _objects(query: 'BaseQuery'):
    o = query._data.class_hierarchies[query._obj]
    def _test(n):
        if n in G:
            return nx.has_path(G, o, n)
        return False
    G = query._data.hierarchy_graph.singular
    singular = set(G.subgraph_view(_test).nodes) - {o}
    G = query._data.hierarchy_graph.plural
    plural = set(G.subgraph_view(_test).nodes) - {o}
    return singular - {o}, plural - {o}

def objects(query: 'BaseQuery', plural=True) -> List[str]:
    from ..data import Data
    if isinstance(query, Data):
        return list(query.plural_hierarchies.keys())
    ss, ps = _objects(query)
    if not plural:
        ps = []
    rels = list(query._data.class_hierarchies[query._obj].relative_names.keys())
    return sorted([s.singular_name.lower() for s in ss] + rels) + sorted([p.plural_name.lower() for p in ps])

def explain(query: 'BaseQuery') -> None:
    from ..data import Data
    if isinstance(query, Data):
        print(f"Database connection {query}")
    if query._obj is None:
        raise TypeError(f"{query} is not a ObjectQuery")
    print(f"{query._obj}:")
    print(f"\t {query._data.class_hierarchies[query._obj].__doc__.strip()}")

def find(query: 'BaseQuery', guess: str, n=10) -> List[str]:
    from ..data import Data
    from .objects import AttributeQuery
    if isinstance(query, Data):
        return query.find_names(guess)
    if isinstance(query, AttributeQuery):
        raise TypeError(f"You cannot search for things in an attribute query. Try using find on an object instead.")
    ss, ps = _objects(query)
    objs = [s.singular_name for s in ss] + [p.plural_name for p in ps]
    factors = [i for s in ss for i in s.products_and_factors] + [query._data.plural_name(i) for p in ps for i in p.products_and_factors]
    sources = objs + factors
    inorder = sorted(sources, key=lambda x: textdistance.jaro_winkler(guess, x), reverse=True)
    return [i.lower() for i in inorder[0:n]]


from astropy.table import Table
from typing import Tuple, Union

from weaveio.readquery.objects import ObjectQuery, AttributeQuery, TableVariableQuery


def join(table: Table, index_column: str,
         object_query: ObjectQuery, join_query: Union[AttributeQuery, str] = None,
         join_on='table') -> Tuple[TableVariableQuery, ObjectQuery]:
    """
    Add each row of an astropy table to a query object.
    The columns of each row will be accessible to all subsequent queries that fork from `object_query`.
    This would traditionally be called a "left join" since only all rows in the table will be returned
    even if there is no match to a weaveio object (null).
    Furthermore, weaveio objects that are not matched will not be returned.

    :param table: Astropy Table containing all data that you want to be accessible to the query
    :param index_column: Column in the table that will be used to join the table to the query object
    :param object_query: Query object that will be joined to the table
    :param join_query: Query object that will be used to join the table to the query object.
                       This can be any attribute query as long as it is singular wrt to `object_query`
                       Leave as None if you want to join the table to the query object using the index_column name as the attribute.
    :param join_on: Type of join to use. "table" is per row in table, "object" is per object in query.
    :return: A tuple of the table variable query and the query object

    Examples:
        Join a table where each row corresponds to a weave_target cname
        >>> table, weave_targets = join(table, 'cname', data.weave_targets.cname)
        `weave_targets` is the subset of `data.weave_targets` that is matched to the table
        `table` is the entire table

        Join a table where each row corresponds to a weave_target cname but is matched to a specific spectrum
        >>> spectra = data.runs[1234].l1single_spectra  # get all spectra belonging to this run
        >>> table, query = join(table, 'cname', spectra, spectra.cname)
        `query` is the subset of `data.runs[1234].l1single_spectra` that is matched to the table
        `table` is the entire table

        Join a table where each row corresponds to an ob and matched based on the first mjd of that OB
        >>> table, query = join(table, 'mjd', data.obs, min(data.obs.exposures.mjd, wrt=data.obs))
        `query` is the subset of `data.obs` that is matched to the table
        `table` is the entire table
    """
    if join_query is None:
        join_query = index_column
    if isinstance(join_query, str):
        join_query = object_query[join_query]
    if not isinstance(join_query, AttributeQuery):
        raise TypeError(f"join_query must be an AttributeQuery or str, not {type(join_query)}")
    G = object_query._G
    param = G.add_parameter(table)
    # filter the input table to only include rows that match the join_query
    applied, applied_var = G.add_apply_to_list(object_query._node, param, '{0}', f'{{0}}.`{index_column}` = {{1}}',
                                               join_query._node, put_null_in_empty=True, parameters=[param])
    applied = G.fold_to_cardinal(applied)
    applied_var = G.G.nodes[applied]['variables'][0]
    row = G.add_unwind_parameter(object_query._node, applied_var, applied)

    ref = G.add_previous_reference(row, object_query._node)
    if join_on == 'table':
        not_null = G.add_scalar_operation(row, '{0} is not null', 'not-null')[0]
        obj_node = G.add_filter(ref, not_null, False)
    elif join_on == 'object':
        obj_node = ref
    else:
        raise ValueError(f"join_on must be either the input table `table` or the input object `object_query`, not {join_on}")
    single = False
    return TableVariableQuery._spawn(object_query, row, '_row', single=True, table=table),\
           ObjectQuery._spawn(object_query, obj_node, object_query._obj, single=single)


import math
from itertools import zip_longest
from typing import Type, Union, Set, List, Tuple
import networkx as nx
from networkx.classes.filters import no_filter

from weaveio.hierarchy import Multiple, Hierarchy, OneOf, Graphable


def normalise_relation(h):
    if not isinstance(h, Multiple):
        h = OneOf(h)
    h.instantate_node()
    relation, node = h, h.node
    return relation, node

def get_all_subclasses(cls: Type) -> List[Type]:
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses

def collapse_classes_to_superclasses(graph, classes):
    """
    Given a list of classes, return the shortest list of classes possible by merging classes together.
    A merge is only possible if they share a base class
    class B(A) and class C(A) can be merged -> class A
    In graph terms, this is equivalent to finding islands (connected components) and returning
     the shallowest node that has a path to all leaves (the head(s))
    """
    G = nx.subgraph_view(graph, filter_node=lambda n: n != Hierarchy and n in classes,
                         filter_edge=lambda *e: graph.edges[e]['type'] == 'subclassed_by')
    for ns in nx.connected_components(G.to_undirected()):
        subgraph = nx.subgraph(G, ns)
        for n in subgraph.nodes:
            if not subgraph.in_degree(n):
                yield n

def hierarchies_from_hierarchy(hier: Type[Hierarchy], done=None, templates=False) -> Set[Type[Hierarchy]]:
    if done is None:
        done = []
    hierarchies = set()
    todo = {h.node if isinstance(h, Multiple) else h for h in hier.parents + hier.children + hier.produces}
    if not templates:
        todo = {h for h in todo if not h.is_template}
    else:
        todo.update({h for h in todo for hh in get_all_class_bases(h) if issubclass(hh, Hierarchy)})
    for new in todo:
        if isinstance(new, Multiple):
            new.instantate_node()
            h = new.node
        else:
            h = new
        if h not in done and h is not hier:
            hierarchies.update(hierarchies_from_hierarchy(h, done, templates))
            done.append(h)
    hierarchies.add(hier)
    return hierarchies


def expand_template_relation(relation):
    """
    Returns a list of relations that relate to each non-template class
    e.g.
    >>> expand_template_relation(Multiple(L1StackedSpectrum))
    [Multiple(L1SingleSpectrum), Multiple(L1StackSpectrum), Multiple(L1SuperstackSpectrum)]
    """
    if not relation.node.is_template:
        return [relation]
    subclasses = [cls for cls in get_all_subclasses(relation.node) if not cls.is_template and issubclass(cls, Hierarchy)]
    return [Multiple(subclass, 0, relation.maxnumber, relation.constrain, relation.relation_idname, relation.one2one) for subclass in subclasses]


def shortest_simple_paths(G, source, target, weight):
    try:
        return nx.shortest_simple_paths(G, source, target, weight)
    except nx.NetworkXNoPath:
        return None

def pick_shortest_multiedge(G, a, b, weight=None):
    edges = set(G.out_edges(a, keys=True)) & set(G.in_edges(b, keys=True))
    if not edges:
        raise nx.NetworkXNoPath("No path found between %s and %s" % (a, b))
    if weight is not None:
        return min(edges, key=lambda e: G.edges[e][weight])
    return edges.pop()

def edges_from_path(G, path, weight=None):
    return list(map(lambda e: pick_shortest_multiedge(G, *e, weight=weight), zip(path[:-1], path[1:])))

def find_forking_path(graph, top, bottom, weight=None):
    done = set()
    parents = nx.subgraph_view(graph, filter_edge=lambda *e: graph.edges[e]['type'] != 'is_a')
    is_a_graph = nx.subgraph_view(graph, filter_edge=lambda *e: graph.edges[e]['type'] == 'is_a')
    gen = nx.all_shortest_paths(graph, bottom, top, weight)
    try:
        todo = [next(gen)]
    except nx.NetworkXNoPath:
        return []
    while todo:
        _path = todo.pop()
        edges = edges_from_path(graph, _path, weight)
        types = [graph.edges[edge]['type'] for edge in edges]
        path = [_path[0]]
        for i, (edge, typ) in enumerate(zip(edges, types)):
            if typ == 'is_a' and edge[1] != top:
                # try to find a path using the superclass
                subpaths = find_forking_path(parents, top, edge[0], weight)
                if subpaths:
                    for subpath in subpaths:
                        done.add((*path[:-1], *subpath))
                else:  # otherwise expand all classes
                    subclasses = [i for i in get_all_subclasses(edge[0]) if not i.is_template]
                    if top in subclasses:
                        inheritance_path = nx.shortest_path(is_a_graph, edge[0], top)
                        done.add((*path, *inheritance_path[1:]))
                    else:
                        for subclass in subclasses:
                            inheritance_path = nx.shortest_path(is_a_graph, edge[0], subclass)
                            for subpath in find_forking_path(graph, top, subclass, weight):
                                done.add((*path, *inheritance_path[1:-1], *subpath))
                break
            else:
                path.append(edge[1])
        else:  # only append here, if we've cycled through all edges without breaking
            done.add(tuple(path))
    return done


def find_paths(graph, a, b, weight='weight') -> Set[Tuple[Type[Hierarchy]]]:
    """
    Returns a set of paths from a to b
    Not all paths are guaranteed to be valid, for example:
        l2superstack.ob yields paths of which `L2superstack-...-L1Single-...-ob` is one.
        This path is invalid, but it is included anyway because it is one of a set
        (through single,stack,superstack,supertarget).
        When queried, invalid paths will yield 0 results so it is not a problem.
    """
    *sorted_nodes, G, use_children = graph.sort_deepest(a, b)
    G = G.reverse()
    _paths = find_forking_path(G, *sorted_nodes, weight)
    _paths = {(p[0], p[0]) if len(p) == 1 else p for p in _paths}
    # put in requested order (a-->b)
    # now remove chains of is_a
    # so x-l1spectrum-l1stacked-l1stack becomes x-l1stack
    reduced_paths = set()
    for ip, path in enumerate(_paths):
        edges = edges_from_path(G, path, weight)
        reduced_path = [edge[0] for edge in edges if G.edges[edge]['type'] not in ['is_a', 'subclassed_by']]
        reduced_path = (*reduced_path, path[-1])
        if not use_children:
            reduced_path = reduced_path[::-1]  # reverse it because the actual database has only parent relations
            edges = (graph.short_edge(e[1], e[0], weight) for e in edges[::-1])
        reduced_paths.add((reduced_path, tuple(edges)))
    return reduced_paths


class HierarchyGraph(nx.MultiDiGraph):
    """
    A multidigraph where edges have keys
    """
    def initialise(self):
        hiers = get_all_subclasses(Hierarchy)
        for h in hiers:
            self._add_hierarchy(h)
        self.remove_node(Hierarchy)
        self._assign_edge_weights()

    def to_neo4j(self, graph):
        node_type = 'AbstractNode'
        for node in self.nodes:
            graph.execute(f"""MERGE (n:{node.__name__}{node_type}:{node_type}) ON CREATE 
            SET n.factors = $factors, n.identifier_builder = $identifier_builder, n.singular_name=$singular_name,
             n.plural_name = $plural_name, n.idname = $idname, n.products = $products""",
                          factors=node.factors, singular_name=node.singular_name, plural_name=node.plural_name,
                          idname=node.idname, identifier_builder=node.identifier_builder, products=node.products)
        for edge in self.edges:
            rel = {k: v for k, v in self.edges[edge].items() if not isinstance(v, Multiple)}
            try:
                rel.update(self.edges[edge]['relation'].to_reldict())
            except KeyError:
                rel['singular'] = False
            if 'constrain' in rel:
                rel['constrain'] = [i.__name__ for i in rel['constrain']]
            graph.execute(f"""MATCH (a:{edge[0].__name__}{node_type}) 
            MATCH (b:{edge[1].__name__}{node_type})
            MERGE (a)-[r:{rel.pop('type')}]->(b)
            ON CREATE SET r = $rel
            """, rel=rel)

    def add_edge(self, u_for_edge, v_for_edge, idname=None, **attr):
        """
        Create an edge if the edge (uniquely identified by attributes) doesn't exist
        the edge key will be an unused integer or the idname of the relation
        if idname is given, it will be used and attr is not checked for uniqueness (only ever 1 edge)
        if idname is not given, attr is checked for uniqueness and key is the number of the same edges
        """
        if not attr:
            return super(HierarchyGraph, self).add_edge(u_for_edge, v_for_edge, idname)
        hsh = hash(tuple(attr.items()))
        if idname is not None:
            # use idname as unique key
            key = super(HierarchyGraph, self).add_edge(u_for_edge, v_for_edge, idname, hsh=hsh, **attr)
            # and make a "general" link as well
            if (u_for_edge, v_for_edge, '*') in self.edges:
                self.edges[u_for_edge, v_for_edge, '*']['singular'] = False
                self.edges[u_for_edge, v_for_edge, '*']['actual_number'] += attr['actual_number']
            else:
                super(HierarchyGraph, self).add_edge(u_for_edge, v_for_edge, '*', hsh=hsh, **attr)
            return key
        if u_for_edge in self and v_for_edge in self:
            edges = set(self.out_edges(u_for_edge, keys=True)) & set(self.in_edges(v_for_edge, keys=True))
        else:
            edges = set()
        if hsh not in (self.edges[e]['hsh'] for e in edges):
            # use an incremented integer as a unique key only if the edge is unique in attr and (u,v)
            return super(HierarchyGraph, self).add_edge(u_for_edge, v_for_edge, hsh=hsh, **attr)

    def surrounding_nodes(self, n):
        return list(self.parents.successors(n)) + list(self.parents.predecessors(n)) + [n]

    def _add_parent(self, child: Type[Hierarchy], parent: Union[Type[Hierarchy], Multiple]):
        relation, parent = normalise_relation(parent)
        relstyle = 'solid' if relation.maxnumber == 1 else 'dashed'
        self.add_edge(child, parent, singular=relation.maxnumber == 1, one2one=relation.one2one,
                       optional=relation.minnumber == 0, style=relstyle, actual_number=relation.maxnumber, type='is_child_of')
        if relation.one2one:
            self.add_edge(parent, child, singular=True, optional=relation.minnumber == 0, style='solid',
                          type='is_parent_of', one2one=relation.one2one,
                          relation=relation, actual_number=relation.maxnumber)
        else:
            self.add_edge(parent, child, singular=False, optional=relation.minnumber == 0, style='dashed',
                          type='is_parent_of', one2one=relation.one2one,
                          relation=relation, actual_number=relation.maxnumber)

    def _add_child(self, parent: Type[Hierarchy], child: Union[Type[Hierarchy], Multiple]):
        relation, child = normalise_relation(child)
        relstyle = 'solid' if relation.maxnumber == 1 else 'dashed'
        self.add_edge(parent, child, singular=relation.maxnumber == 1, one2one=relation.one2one,
                       optional=relation.minnumber == 0, type='is_parent_of',
                       relation=relation, style=relstyle, actual_number=relation.maxnumber)
        self.add_edge(child, parent, singular=True, optional=relation.minnumber == 0,  one2one=relation.one2one,
                      style='solid', type='is_child_of', actual_number=relation.maxnumber)

    def _add_self_reference(self, relation):
        relation, h = normalise_relation(relation)
        relstyle = 'solid' if relation.maxnumber == 1 else 'dashed'
        self.add_edge(h, h, singular=relation.maxnumber == 1, optional=relation.minnumber == 0,
                      one2one=relation.one2one, style=relstyle, actual_number=relation.maxnumber, type='is_parent_of')

    def _add_inheritance(self, hierarchy, base):
        self.add_edge(base, hierarchy, type='subclassed_by', actual_number=math.inf, style='dotted', optional=False,
                      one2one=False)
        self.add_edge(hierarchy, base, type='is_a', actual_number=0, style='dotted', optional=False, one2one=False)

    def _assign_edge_weights(self):
        for u, v, k, d in self.edges(data=True, keys=True):
            if d['type'] == 'is_a':
                weight = 0
            elif d['type'] == 'subclassed_by':
                weight = math.inf
            elif d['singular']:
                weight = 0
            else:
                weight = d['actual_number']
            self.edges[u, v, k]['weight'] = weight

    def _add_hierarchy(self, hierarchy: Type[Hierarchy]):
        """
        For a given hierarchy, traverse all its required inputs (parents and children)
        """
        self.add_node(hierarchy)
        for parent in hierarchy.parents:
            if normalise_relation(hierarchy)[1] is parent:
                self._add_self_reference(parent)
            else:
                self._add_parent(hierarchy, parent)
        for child in hierarchy.children:
            if normalise_relation(hierarchy)[1] is child:
                self._add_self_reference(child)
            else:
                self._add_child(hierarchy, child)
        for inherited in hierarchy.__bases__:
            if issubclass(inherited, Hierarchy):
                self._add_inheritance(hierarchy, inherited)

    def ancestor_subgraph(self, source):
        nodes = nx.ancestors(self.parents_and_subclassed_by, source)
        nodes.add(source)
        return nx.subgraph(self, nodes).copy()

    def subgraph_view(self, filter_node=no_filter, filter_edge=no_filter) -> 'HierarchyGraph':
        return nx.subgraph_view(self, filter_node, filter_edge)  # type: HierarchyGraph

    @property
    def inheritance(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['type'] == 'is_a').copy()

    @property
    def parents_and_subclassed_by(self):
        allowed = ['is_parent_of', 'subclassed_by']
        return self.subgraph_view(filter_edge=lambda *e: any(i == self.edges[e]['type'] for i in allowed)).copy()

    @property
    def parents(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['type'] == 'is_parent_of').copy()

    @property
    def children(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['type'] == 'is_child_of').copy()

    @property
    def parents_and_inheritance(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['type'] == 'is_parent_of' or 'is_a' == self.edges[e]['type']).copy()

    @property
    def children_and_inheritance(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['type'] == 'is_child_of' or 'is_a' == self.edges[e]['type']).copy()

    @property
    def traversal(self):
        def func(u, v):
            edge = self.edges[u, v]
            typ = edge['type']
            return (typ == 'is_parent_of') or (typ == 'is_a') or edge['one2one']
        return self.subgraph_view(filter_edge=func)

    @property
    def nonoptional(self):
        return self.subgraph_view(filter_edge=lambda *e: not self.edges[e]['optional']).copy()

    def sort_deepest(self, a, b):
        """Returns [a, b] or [b, a], in order of increasing depth in the graph"""
        if a is b:
            return a, b, self.parents_and_inheritance, False
        if a not in self:
            raise nx.NodeNotFound(f"Node {a} not found in graph {self}")
        if b not in self:
            raise nx.NodeNotFound(f"Node {b} not found in graph {self}")
        if b in nx.ancestors(self.parents_and_inheritance, a):
            return b, a, self.parents_and_inheritance, False
        elif a in nx.ancestors(self.parents_and_inheritance, b):
            return a, b, self.parents_and_inheritance, False
        elif a in nx.ancestors(self.children_and_inheritance, b):
            return a, b, self.children_and_inheritance, True
        elif b in nx.ancestors(self.children_and_inheritance, a):
            return b, a, self.children_and_inheritance, True
        else:
            raise nx.NetworkXNoPath(f"There is no path between {a} and {b}")

    def short_edge(self, a, b, weight=None):
        """return the shortest edge in the multidigraph for a given digraph edge """
        return pick_shortest_multiedge(self, a, b, weight)

    @property
    def singular(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['weight'] == 0)

    @property
    def plural(self):
        return self.subgraph_view(filter_edge=lambda *e: self.edges[e]['weight'] > 0)

    def find_paths(self, a, b, singular):
        """
        Returns a set of paths from a to b
        Not all paths are guaranteed to be valid, for example:
            l2superstack.ob yields paths of which `L2superstack-...-L1Single-...-ob` is one.
            This path is invalid, but it is included anyway because it is one of a set
            (through single,stack,superstack,supertarget).
            When queried, invalid paths will silently yield 0 results so it is not a problem.
        """
        try:
            return find_paths(self.singular.nonoptional, a, b)
        except nx.NetworkXNoPath:
            try:
                return find_paths(self.singular, a, b)
            except nx.NetworkXNoPath as e:
                if singular:
                    raise e
                try:
                    return find_paths(self.nonoptional, a, b)
                except nx.NetworkXNoPath:
                    return find_paths(self, a, b)

    def edge_weights(self, path) -> List[int]:
        return [self.edges[self.short_edge(*e, 'weight')]['weight'] for e in nx.utils.pairwise(path)]

    def edge_path_is_singular(self, edges):
        return not any(self.edges[e]['weight'] > 0 for e in edges)

    def path_is_singular(self, path) -> bool:
        return not any(e > 0 for e in self.edge_weights(path))


def get_all_class_bases(cls: Type[Graphable]) -> Set[Type[Graphable]]:
    new = set()
    for b in cls.__bases__:
        if b is Graphable or not issubclass(b, Graphable):
            continue
        new.add(b)
        new.update(get_all_class_bases(b))
    return new



from weaveio.data import Data as BaseData
from weaveio.opr3.l1files import RawFile, L1SingleFile, L1StackFile, L1SuperstackFile, L1SupertargetFile
from weaveio.opr3.l2files import L2StackFile, L2SuperstackFile, L2SingleFile, L2SupertargetFile


class Data(BaseData):
    filetypes = [
        RawFile,
        L1SingleFile, L2SingleFile,
        L1StackFile, L2StackFile,
        L1SuperstackFile, L2SuperstackFile,
        L1SupertargetFile, L2SupertargetFile
]


from .objects import Query
from .functions import *
from .aggregations import *
from .results import filtered, masked, apply, Table
from .uploads import join
from .helpers import *
from .split import split
from .align import align



from functools import lru_cache
from typing import List, Tuple

import graphviz
import networkx as nx
from networkx.classes.graphviews import generic_graph_view
from networkx.drawing.nx_pydot import to_pydot


class HashedDiGraph(nx.DiGraph):
    hash_edge_attr = 'type'
    hash_node_attr = 'i'

    @property
    def name(self) -> str:
        return nx.weisfeiler_lehman_graph_hash(self, edge_attr=self.hash_edge_attr, node_attr=self.hash_node_attr)

    @name.setter
    def name(self, s):
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(self.name)

    def add_node(self, node_for_adding, **attr):
        """add_node but return existing node if it is matched, ignoring attributes"""
        if node_for_adding in self.nodes:
            return node_for_adding
        return super().add_node(node_for_adding, **attr)

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        """add_edge but return existing node if it is matched"""
        if (u_of_edge, v_of_edge) in self.edges:
            if self.edges[(u_of_edge, v_of_edge)] == attr:
                return (u_of_edge, v_of_edge)
        super().add_edge(u_of_edge, v_of_edge, **attr)


def plot_graph(graph, highlight_nodes=None, highlight_edges=None):
    highlight_nodes = [] if highlight_nodes is None else highlight_nodes
    highlight_edges = [] if highlight_edges is None else highlight_edges
    g = nx.DiGraph()
    for n in graph.nodes:
        g.add_node(n)
        if n in highlight_nodes:
            g.nodes[n]['style'] = 'filled'
    for e in graph.edges():
        try:
            label = e['statement'].make_cypher(graph.nodes)
        except:
            label = ''
        g.add_edge(*e, **graph.edges[e], label=label)
        if e in highlight_edges:
            g.edges[e]['arrowsize'] = 2
    nx.relabel_nodes(g, {n: f"{d['i']}\n{d['label']}" for n, d in graph.nodes(data=True)}, copy=False)
    return graphviz.Source(to_pydot(g).to_string())


def make_node(graph: HashedDiGraph, parent, type: str, statement: 'Statement', single, **edge_data):
    i = graph.number_of_nodes()
    try:
        label = str(statement.output_variables[0])
    except (IndexError, AttributeError):
        label = 'node'
    node = i
    for a, b, d in graph.out_edges(parent, data=True):
        if d.get('statement', None) == statement:
            return b  # node is already there so return it
    graph.add_node(node, label=label, i=i, variables=[])
    if parent is not None:
        graph.add_edge(parent, node, type=type, statement=statement, single=single, **edge_data)
    if statement is not None:
        statement.edge = (parent, node)
    return node


def add_start(graph: HashedDiGraph, name):
    return make_node(graph, None, name, None, False)


def add_traversal(graph: HashedDiGraph, parent, statement, single=False, unwound=None):
    if unwound is not None:
        parent, deps = unwound, [parent]
    else:
        parent, deps = parent, []
    n = make_node(graph, parent, 'traversal', statement, single)
    for d in deps:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def add_node_reference(graph: HashedDiGraph, statement, parent, *deps):
    n = make_node(graph, parent, 'traversal', statement, single=True)
    for d in deps:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def add_filter(graph: HashedDiGraph, parent, dependencies, statement, force_single=False, split_node=False):
    if force_single:
        single = True
    else:
        single = graph.edges[list(graph.in_edges(parent))[0]].get('single', False)
    n = make_node(graph, parent, 'filter', statement, single, split_node=split_node)
    for d in dependencies:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def add_aggregation(graph: HashedDiGraph, parent, wrt, statement, type='aggr', single=False, dependencies=None):
    dependencies = [] if dependencies is None else dependencies
    n = make_node(graph, parent, type, statement, single)
    graph.add_edge(n, wrt, type='wrt', style='dashed')
    for d in dependencies:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def add_operation(graph: HashedDiGraph, parent, dependencies, statement):
    n = make_node(graph, parent, 'operation', statement, True)
    for d in dependencies:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def add_return(graph: HashedDiGraph, index, columns, statement):
    n = make_node(graph, index, 'return', statement, True)
    for d in columns:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def add_unwind(graph: HashedDiGraph, parent, statement, *dependencies):
    n = make_node(graph, parent, 'traversal', statement, single=False)
    for d in dependencies:
        graph.add_edge(d, n, type='dep', style='dotted')
    return n


def subgraph_view(graph: HashedDiGraph, excluded_edge_type=None, only_edge_type=None,
                  only_nodes: List = None, excluded_nodes: List = None,
                  only_edges: List[Tuple] = None, excluded_edges: List[Tuple] = None,
                  path_to = None,
                  ) -> HashedDiGraph:
    """
    filters out edges and nodes
    """
    excluded_edges = set([] if excluded_edges is None else excluded_edges)
    excluded_nodes = set([] if excluded_nodes is None else excluded_nodes)
    if excluded_edge_type is not None:
        excluded_edges |= {e for e in graph.edges if graph.edges[e].get('type', '') == excluded_edge_type}
    if only_edge_type is not None:
        excluded_edges |= {e for e in graph.edges if graph.edges[e].get('type', '') != only_edge_type}
    if only_nodes is not None:
        excluded_nodes |= {n for n in graph.nodes if n not in only_nodes}
    if only_edges is not None:
        excluded_edges |= {e for e in graph.edges if e not in only_edges}
    r = nx.restricted_view(graph, excluded_nodes, excluded_edges)  # type: HashedDiGraph
    if path_to:
        r = nx.subgraph_view(r, lambda n:  nx.has_path(graph, n, path_to))
    return r


def partial_reverse(G: nx.DiGraph, edges: List[Tuple]) -> HashedDiGraph:
    newG = generic_graph_view(G).copy()
    succ = {}
    pred = {}
    for node in newG.nodes:
        succs = {succ: data for succ, data in newG._succ[node].items() if (node, succ) not in edges}
        succs.update({pred: data for pred, data in newG._pred[node].items() if (pred, node) in edges})
        preds = {pred: data for pred, data in newG._pred[node].items() if (pred, node) not in edges}
        preds.update({succ: data for succ, data in newG._succ[node].items() if (node, succ) in edges})
        succ[node] = succs
        pred[node] = preds
    newG._succ = succ
    newG._pred = pred
    newG._adj = newG._succ
    return newG


@lru_cache()
def get_above_state_traversal_graph(graph: HashedDiGraph):
    """
    traverse backwards and only follow wrt when there is a choice
    """
    wrt_graph = subgraph_view(graph, only_edge_type='wrt')
    def allowed_traversal(a, b):
        if graph.edges[(a, b)]['type'] == 'wrt':
            return True
        if graph.edges[(a, b)]['type'] == 'dep':
            return False
        if wrt_graph.out_degree(b):
            return False  # can't traverse other edges if there is a wrt
        return True
    alloweds = nx.subgraph_view(graph, filter_edge=allowed_traversal)  # type: HashedDiGraph
    return partial_reverse(alloweds, [(a, b) for a, b, data in graph.edges(data=True) if data['type'] != 'wrt'])


def node_dependencies(graph, node):
    dag = subgraph_view(graph, excluded_edge_type='wrt')
    return {n for n in graph.nodes if nx.has_path(dag, n, node)} - {node}



import pytest
from weaveio import *


def test_rerun(data):
    l2s = data.l2stacks
    l2s = l2s[any(l2s.fibre_target.surveys == '/WL.*/', wrt=l2s)]
    l2s = l2s[l2s['ha_6562.80_flux'] > 0]
    ratio = l2s['[oiii]_5006.77_flux'] / l2s['ha_6562.80_flux']
    max(ratio)
    ratio = l2s['[oiii]_5006.77_flux'] / l2s['ha_6562.80_flux']
    one_l2 = l2s[max(ratio) == ratio]
    t = one_l2[['[oiii]_5006.77_flux', 'ha_6562.80_flux', 'cname']]
    cname1 = t()['cname'][0]
    l2s = l2s[l2s['[oiii]_5006.77_flux'] > 0]
    ratio = l2s['[oiii]_5006.77_flux'] / l2s['ha_6562.80_flux']
    one_l2 = l2s[max(ratio) == ratio]
    t = one_l2[['[oiii]_5006.77_flux', 'ha_6562.80_flux', 'cname']]
    cname2 = t()['cname'][0]
    assert cname2 == cname1


def test_rerun_with_limit(data):
    obs = data.obs.id
    assert np.all(obs(limit=10) == obs(limit=10))


def test_merge_tables(data):
    def noise_spectra_query(parent, camera, targuse='S', split_into_subqueries=True):
        if split_into_subqueries:
            parent = split(parent)
        stacks = parent.l1stack_spectra[(parent.l1stack_spectra.targuse == targuse) & (parent.l1stack_spectra.camera == camera)]
        singles = stacks.l1single_spectra
        singles_table = singles[['flux', 'ivar']]
        query = stacks[[stacks.ob.id, {'stack_flux': 'flux', 'stack_ivar': 'ivar'}, 'wvl', {'single_': singles_table}]]
        return query

    for index, query in noise_spectra_query(data.obs, 'red'):
        t = query(limit=1)
        assert t['id'] == index
        assert len(t['stack_flux'].shape) == 1
        assert len(t['single_flux'].shape) == 2
        break


def test_aggregations_of_aggregations_of_aggregations(data):
    obs = data.obs
    runs = obs.runs
    single = runs.l1single_spectra
    spec = single.l1stack_spectra
    single_mean = mean(spec.snr, wrt=single)
    run_mean = mean(single_mean, wrt=runs)
    ob_mean = mean(run_mean, wrt=obs)
    assert len(ob_mean()) == count(obs)()


def test_noss_access(data):
    parent = data.obs
    stacks = parent.l1stack_spectra[(parent.l1stack_spectra.targuse == 'S') & (parent.l1stack_spectra.camera == 'red')]
    singles = stacks.l1single_spectra  # get single spectra for each stack spectrum
    ss_query = stacks[['ob', {'stack_flux': 'flux', 'stack_ivar': 'ivar'}, 'wvl', {'single_': singles[['flux', 'ivar', 'wvl']]}]]
    noss_query = stacks.noss[['ob', {'stack_flux': 'flux', 'stack_ivar': 'ivar'}, 'wvl', {'single_': singles.noss[['flux', 'ivar', 'wvl']]}]]
    ss = ss_query(limit=10)
    noss = noss_query(limit=10)
    assert all(ss['ob'] == noss['ob'])
    assert all(np.max(ss['stack_flux'], axis=1) <= np.max(noss['stack_flux'], axis=1))


import logging
import sys
from pathlib import Path
from typing import Union, List, Tuple, Dict

import inspect
from astropy.io import fits
from astropy.io.fits.hdu.base import _BaseHDU
from astropy.table import Table as AstropyTable
import numpy as np
import pandas as pd

from weaveio.config_tables import progtemp_config
from weaveio.file import File, PrimaryHDU, TableHDU, BinaryHDU
from weaveio.hierarchy import unwind, collect, Multiple, Hierarchy, OneOf, Optional
from weaveio.opr3.hierarchy import Survey, Targprog, Catalogue, \
    WeaveTarget, SurveyTarget, Fibre, FibreTarget, Progtemp, ArmConfig, Obstemp, \
    OBSpec, OB, Exposure, Run, CASU, RawSpectrum, _predicate, WavelengthHolder, System
from weaveio.opr3.l1 import L1SingleSpectrum, L1StackSpectrum, L1SuperstackSpectrum, L1SupertargetSpectrum, L1Spectrum, L1StackedSpectrum, NoSS
from weaveio.writequery import groupby, CypherData


def explode_pandas_series(df: pd.DataFrame, column: str) -> pd.DataFrame:
    copy = df.copy()
    del copy[column]
    return df.apply(lambda x: pd.Series(x[column]), axis=1).stack().reset_index(level=1, drop=1).to_frame(column).join(copy)

def groupby_aggregate_pandas_list_column(df: pd.DataFrame, groupby: Union[str, List[str]], list_column: str) -> pd.DataFrame:
    # must be sorted because weaveio is sensitive to ordering!
    return df.groupby(groupby)[list_column].agg(lambda x: tuple(sorted({j for i in x for j in i}))).reset_index()

class HeaderFibinfoFile(File):
    is_template = True

    @classmethod
    def length(cls, path, part=None):
        return fits.open(path)['FIBTABLE'].header['NAXIS2']

    @classmethod
    def read_fibinfo_dataframe(cls, path, slc=None):
        hdus = fits.open(path)
        fibinfo_hdu = hdus['FIBTABLE']
        fibinfo = AstropyTable(fibinfo_hdu.data).to_pandas()
        fibinfo.columns = [i.lower() for i in fibinfo.columns]
        fibinfo.rename(columns=lambda x: f'{x[1:]}_error' if x.startswith('emag') else x, inplace=True)
        if 'nspec' in fibinfo.columns:
            fibinfo['spec_index'] = fibinfo['nspec'] - 1
        slc = slice(None) if slc is None else slc
        selected = fibinfo.iloc[slc]
        t = selected[~selected['fibreid'].isnull()]
        t['targsrvy'] = t.targsrvy.apply(str)
        t['targprog'] = t.targprog.apply(str).str.split('|').apply(tuple)
        t['targcat'] = t.targcat.apply(str)
        if np.any(t.targsrvy.apply(str).str.split(',').apply(len) > 1):
            raise ValueError(f"There are multiple surveys defined in targsrvy. The schema needs to be revised")
        return t.drop_duplicates()

    @classmethod
    def read_surveyinfo(cls, df_fibinfo):
        df_svryinfo = df_fibinfo[['targsrvy', 'targprog', 'targcat']].drop_duplicates()
        return groupby_aggregate_pandas_list_column(df_svryinfo, ['targsrvy', 'targcat'], 'targprog')

    @classmethod
    def read_fibretargets(cls, df_svryinfo, df_fibinfo, obspec):
        fibinfo = CypherData(df_fibinfo)
        with unwind(fibinfo) as fibrow:
            fibre = Fibre(id=fibrow['fibreid'])  #  must be up here for some reason otherwise, there will be duplicates
            survey = Survey.find(name=fibrow['targsrvy'])
            catalogue = Catalogue.find(name=fibrow['targcat'], survey=survey)
            with unwind(fibrow['targprog']) as targprog:
                targprog = Targprog.find(name=targprog, survey=survey)
            targprogs = collect(targprog)
            weavetarget = WeaveTarget(cname=fibrow['cname'])
            surveytarget = SurveyTarget(catalogue=catalogue, weave_target=weavetarget, targprogs=targprogs, tables=fibrow)
            fibtarget = FibreTarget(survey_target=surveytarget, fibre=fibre, obspec=obspec, tables=fibrow)
        return collect(fibtarget, fibrow)

    @classmethod
    def read_distinct_survey_info(cls, df_svryinfo):
        # must be sorted because weaveio is sensitive to ordering!
        survey_rows = CypherData(groupby_aggregate_pandas_list_column(df_svryinfo, 'targsrvy', 'targprog'), 'survey_rows')
        with unwind(survey_rows) as survey_row:  # unique survey
            survey = Survey(name=survey_row['targsrvy'])
            with unwind(survey_row['targprog']) as targprog_name:
                targprog = Targprog(survey=survey, name=targprog_name)
            collect(targprog)
        survey_list = collect(survey)
        survey_dict = groupby(survey_list, 'name')

        cat_rows = CypherData(df_svryinfo, 'targcat_rows')
        with unwind(cat_rows) as cat_row:  # unique cat
            survey = survey_dict[cat_row['targsrvy']]
            with unwind(cat_row['targprog']) as targprog_name:
                targprog = Targprog.find(survey=survey, name=targprog_name)
            targprogs = collect(targprog)
            catalogue = Catalogue(name=cat_row['targcat'], survey=survey, targprogs=targprogs)
        catalogue_list = collect(catalogue)
        targprog_rows = CypherData(explode_pandas_series(df_svryinfo, 'targprog'), 'targprog_rows')
        with unwind(targprog_rows) as targprog_row:  # unique targprog
            targprog = Targprog.find(survey=survey_dict[targprog_row['targsrvy']], name=targprog_row['targprog'])
        targprog_list = collect(targprog)
        return survey_list, targprog_list, catalogue_list

    @classmethod
    def read_hierarchy(cls, header, df_svryinfo):
        surveys, subprogrammes, catalogues = cls.read_distinct_survey_info(df_svryinfo)
        runid = int(header['RUN'])
        camera = str(header['CAMERA'].lower()[len('WEAVE'):])
        obstart = np.round(float(header['OBSTART']), 6)
        res = str(header['VPH']).rstrip('123').lower().replace('res', '')
        obtitle = str(header['OBTITLE'])
        xml = str(header['CAT-NAME']).replace('./', '') if 'xml' in header['CAT-NAME'] else str(header['INFILE'])  # stack files do this
        obid = float(header['OBID'])
        casu = CASU(id=header['CASUID'])
        sys = System(version=header['sysver'])
        progtemp = Progtemp.from_progtemp_code(header['PROGTEMP'])
        vph = int(progtemp_config[(progtemp_config['mode'] == progtemp.instrument_configuration.mode)
                                  & (progtemp_config['resolution'] == res.lower().replace('res', ''))][f'{camera}_vph'].iloc[0])
        arm = ArmConfig(vph=vph, resolution=res, camera=camera)  # must instantiate even if not used
        obstemp = Obstemp.from_header(header)
        obspec = OBSpec(xml=xml, title=obtitle, obstemp=obstemp, progtemp=progtemp,
                        catalogues=catalogues, targprogs=subprogrammes, surveys=surveys)
        ob = OB(id=obid, mjd=obstart, obspec=obspec)
        exposure = Exposure.from_header(ob, header)
        adjunct_run = Run.find(anonymous_parents=[exposure], exclude=[Run.find(id=runid)])
        run = Run(id=runid, arm_config=arm, exposure=exposure, adjunct=adjunct_run)
        return {'ob': ob, 'obspec': obspec, 'arm_config': arm, 'exposure': exposure, 'run': run, 'adjunct_run': adjunct_run, 'casu': casu, 'system': sys}

    @classmethod
    def read_schema(cls, path: Path, slc: slice = None):
        header = cls.read_header(path)
        fibinfo = cls.read_fibinfo_dataframe(path, slc)
        srvyinfo = cls.read_surveyinfo(fibinfo)
        hiers = cls.read_hierarchy(header, srvyinfo)
        fibretarget_collection, fibrows = cls.read_fibretargets(srvyinfo, fibinfo, hiers['obspec'])
        return hiers, header, fibinfo, fibretarget_collection, fibrows

    @classmethod
    def read_header(cls, path: Path, i=0):
        return fits.open(path)[i].header

    @classmethod
    def read_fibtable(cls, path: Path):
        return AstropyTable(fits.open(path)[cls.fibinfo_i].data)

    @classmethod
    def read(cls, directory: Path, fname: Path, slc: slice = None, part=None) -> 'File':
        raise NotImplementedError


class RawFile(HeaderFibinfoFile):
    match_pattern = 'r[0-9]+\.fit'
    hdus = {'primary': PrimaryHDU, 'counts1': BinaryHDU, 'counts2': BinaryHDU,
            'fibtable': TableHDU, 'guidinfo': TableHDU, 'metinfo': TableHDU}
    parents = [OneOf(RawSpectrum, one2one=True)]
    produces = [CASU, System]  # extra things which are not necessarily children of this object, cannot include parents
    children = [Optional('self', idname='adjunct', one2one=True), CASU, System]

    @classmethod
    def fname_from_runid(cls, runid):
        return f'r{runid:07.0f}.fit'

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None, part=None):
        path = Path(directory) / Path(fname)
        hiers, header, fibinfo, fibretarget_collection, fibrow_collection = cls.read_schema(path, slc)
        adjunct_run = hiers['adjunct_run']
        adjunct_raw = RawSpectrum.find(anonymous_parents=[adjunct_run])
        adjunct_rawfile = RawFile.find(anonymous_children=[adjunct_raw])
        raw = RawSpectrum(run=hiers['run'], adjunct=adjunct_raw)
        hdus, file, _ = cls.read_hdus(directory, fname, raw_spectrum=raw, casu=hiers['casu'],
                                      system=hiers['system'], adjunct=adjunct_rawfile)
        where = {'counts1': 1, 'counts2': 2}
        for product in raw.products:
            raw.attach_product(product, hdus[where[product]])
        return file


class L1File(HeaderFibinfoFile):
    singular_name = 'l1file'
    is_template = True
    hdus = {'primary': PrimaryHDU, 'flux': BinaryHDU, 'ivar': BinaryHDU,
            'flux_noss': BinaryHDU, 'ivar_noss': BinaryHDU,
            'sensfunc': BinaryHDU, 'fibtable': TableHDU}
    children = [Optional('self', idname='adjunct', one2one=True), WavelengthHolder, CASU, System]
    parents = [Multiple(L1Spectrum, maxnumber=1000, one2one=True)]
    produces = [CASU, System, NoSS]  # extra things which are not necessarily children of this object, cannot include parents

    @classmethod
    def wavelengths(cls, rootdir: Path, fname: Union[Path, str]):
        hdulist = fits.open(rootdir / fname)
        header = hdulist[1].header
        increment, zeropoint, size = header['cd1_1'], header['crval1'], header['naxis1']
        return WavelengthHolder(wvl=(np.arange(0, size) * increment) + zeropoint,
                                cd1_1=header['cd1_1'], crval1=header['crval1'], naxis1=header['naxis1'])

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str],
                  **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[int, 'HDU'], 'File', List[_BaseHDU]]:
        return super().read_hdus(directory, fname, **hierarchies, wavelength_holder=cls.wavelengths(directory, fname))

    @classmethod
    def attach_products_to_spectrum(cls, spectrum, index, hdus, names):
        for product in spectrum.products:
            spectrum.attach_product(product, hdus[names[product]], index=index)


class L1SingleFile(L1File):
    singular_name = 'l1single_file'
    match_pattern = 'single_\d+\.fit'
    parents = [OneOf(RawFile, one2one=True), Multiple(L1SingleSpectrum, maxnumber=1000, one2one=True)]
    children = [Optional('self', idname='adjunct', one2one=True), WavelengthHolder, CASU, System]

    @classmethod
    def fname_from_runid(cls, runid):
        return f'single_{runid:07.0f}.fit'

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None, part=None):
        fname = Path(fname)
        directory = Path(directory)
        absolute_path = directory / fname
        hiers, header, fibinfo, fibretarget_collection, fibrow_collection = cls.read_schema(absolute_path, slc)
        casu = hiers['casu']
        system = hiers['system']
        inferred_raw_fname = fname.with_name(fname.name.replace('single_', 'r'))
        matched_files = list(directory.rglob(inferred_raw_fname.name))
        if not matched_files:
            logging.warning(f"{fname} does not have a matching raw file. "
                            f"The database will create the link but the rawspectra will not be available until the missing file is.")
            inferred_raw_fname = inferred_raw_fname.name
        elif len(matched_files) > 1:
            raise FileExistsError(f"Whilst searching for {inferred_raw_fname}, we found more than one match:"
                                  f"\n{matched_files}")
        else:
            inferred_raw_fname = str(matched_files[0].name)
        adjunct_run = hiers['adjunct_run']
        adjunct_raw = RawSpectrum.find(anonymous_parents=[adjunct_run])
        adjunct_raw_file = RawFile.find(anonymous_parents=[adjunct_raw])
        adjunct_file = cls.find(anonymous_parents=[adjunct_raw_file])
        raw = RawSpectrum(run=hiers['run'])
        rawfile = RawFile(fname=inferred_raw_fname, casu=casu, system=hiers['system'], raw_spectrum=raw, adjunct=adjunct_raw_file, path='<MISSING>')  # merge this one instead of finding, then we can start from single or raw files
        wavelengths = cls.wavelengths(directory, fname)
        with unwind(fibretarget_collection, fibrow_collection) as (fibretarget, fibrow):
            adjunct = L1SingleSpectrum.find(nspec=fibrow['nspec'], raw_spectrum=adjunct_raw, fibre_target=fibretarget)
            adjunct_noss = NoSS.find(anonymous_parents=[adjunct])
            single_spectrum = L1SingleSpectrum(nspec=fibrow['nspec'], raw_spectrum=raw, fibre_target=fibretarget,
                                               wavelength_holder=wavelengths, arm_config=hiers['arm_config'],
                                               tables=fibrow, adjunct=adjunct)
            noss = NoSS(adjunct=adjunct_noss, l1_spectrum=single_spectrum)
        single_spectra, nosses, fibrows = collect(single_spectrum, noss, fibrow)
        hdus, file, _ = cls.read_hdus(directory, fname, raw_file=rawfile, adjunct=adjunct_file,
                                      l1single_spectra=single_spectra,
                                      casu=casu, system=system)
        with unwind(single_spectra, nosses, fibrows) as (single_spectrum, noss, fibrow):
            spec = L1SingleSpectrum.from_cypher_variable(single_spectrum)
            noss = NoSS.from_cypher_variable(noss)
            cls.attach_products_to_spectrum(spec, fibrow['spec_index'], hdus, {'flux': 1, 'ivar': 2, 'sensfunc': 5})
            cls.attach_products_to_spectrum(noss, fibrow['spec_index'], hdus, {'flux': 3, 'ivar': 4, 'sensfunc': 5})
        single_spectra = collect(single_spectrum)  # must collect at the end
        return file


class L1StackedFile(L1File):
    singular_name = 'l1stacked_file'
    is_template = True
    SpectrumType = None

    @classmethod
    def parent_runids(cls, path):
        header = cls.read_header(path, 0)
        runids = [int(v) for k, v in header.items() if k.startswith('RUNS0')]
        if len(runids) == 0:
            runids = [int(v.split('.')[0].split('_')[1]) for k, v in header.items() if k.startswith('PROV0') and 'formed' not in v]
        return runids

    @classmethod
    def fname_from_runid(cls, runid):
        return f'stack_{runid:07.0f}.fit'

    @classmethod
    def get_single_files_fnames(cls, directory: Path, fname: Path):
        runids = cls.parent_runids(directory / fname)
        fnames = [L1SingleFile.fname_from_runid(runid) for runid in runids]
        assert len(fnames) > 1, f'{fname} doesnt have more than one l1single runid'
        return fnames

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None, part=None):
        """
        L1Stack inherits everything from the lowest numbered single/raw files so we are missing data,
        therefore we require that all the referenced Runs are present before loading in
        """
        fname = Path(fname)
        directory = Path(directory)
        hiers, header, fibinfo, fibretarget_collection, fibrow_collection = cls.read_schema(directory / fname, slc)
        ob = hiers['ob']
        armconfig = hiers['arm_config']
        # unwind all single files that went into making this stack
        single_fnames = cls.get_single_files_fnames(directory, fname)
        with unwind(CypherData(single_fnames)) as single_fname:
            single_file = L1SingleFile.find(fname=single_fname)
            adjunct_single_file = L1SingleFile.find(anonymous_parents=[single_file], exclude=[single_file])
        single_files, adjunct_single_files = collect(single_file, adjunct_single_file)
        adjunct_file = cls.find(anonymous_parents=[adjunct_single_files])
        wavelengths = cls.wavelengths(directory, fname)
        with unwind(fibretarget_collection, fibrow_collection) as (fibretarget, fibrow):
            adjunct = cls.SpectrumType.find(nspec=fibrow['nspec'], anonymous_parents=[fibretarget], anonymous_children=[adjunct_file])
            adjunct_noss = NoSS.find(anonymous_parents=[adjunct])
            with unwind(single_files) as single_file:
                single_spectrum = L1SingleSpectrum.find(nspec=fibrow['nspec'], anonymous_parents=[fibretarget], anonymous_children=[single_file])
            single_spectra = collect(single_spectrum)
            # use the generic "L1stackspectrum" to avoid writing out again for superstacks
            stack_spectrum = cls.SpectrumType(l1single_spectra=[single_spectra[i] for i, _ in enumerate(single_fnames)], ob=ob,
                                              arm_config=armconfig, fibre_target=fibretarget,
                                              tables=fibrow, nspec=fibrow['nspec'],
                                              adjunct=adjunct, wavelength_holder=wavelengths)
            noss = NoSS(adjunct=adjunct_noss, l1_spectrum=stack_spectrum)
        stack_spectra, nosses, fibrows = collect(stack_spectrum, noss, fibrow)
        d = {cls.SpectrumType.plural_name: stack_spectra}
        hdus, file, _ = cls.read_hdus(directory, fname, l1single_files=single_files,
                                      adjunct=adjunct_file, **d, **hiers)
        with unwind(stack_spectra, nosses, fibrows) as (stack_spectrum, noss, fibrow):
            spec = cls.SpectrumType.from_cypher_variable(stack_spectrum)
            noss = NoSS.from_cypher_variable(noss)
            cls.attach_products_to_spectrum(spec, fibrow['spec_index'], hdus, {'flux': 1, 'ivar': 2, 'sensfunc': 5})
            cls.attach_products_to_spectrum(noss, fibrow['spec_index'], hdus, {'flux': 3, 'ivar': 4, 'sensfunc': 5})
        stack_spectra = collect(stack_spectrum)  # must collect at the end
        return file


class L1StackFile(L1StackedFile):
    singular_name = 'l1stack_file'
    match_pattern = 'stack_[0-9]+\.fit'
    parents = [Multiple(L1SingleFile, maxnumber=10, constrain=(OB, ArmConfig)),
               Multiple(L1StackSpectrum, maxnumber=1000, one2one=True)]
    children = [Optional('self', idname='adjunct', one2one=True), WavelengthHolder, CASU, System]
    SpectrumType =  L1StackSpectrum

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str], **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[int, 'HDU'], 'File', List[_BaseHDU]]:
        return super().read_hdus(directory, fname, casu=hierarchies['casu'], system=hierarchies['system'],
                                 l1single_files=hierarchies['l1single_files'],
                                 ob=hierarchies['ob'], arm_config=hierarchies['arm_config'],
                                 l1stack_spectra=hierarchies['l1stack_spectra'], adjunct=hierarchies['adjunct'])


class L1SuperstackFile(L1StackedFile):
    singular_name = 'l1superstack_file'
    match_pattern = 'superstack_[0-9]+\.fit'
    parents = [Multiple(L1SingleFile, maxnumber=5, constrain=(OBSpec, ArmConfig)),
               Multiple(L1SuperstackSpectrum, maxnumber=1000, one2one=True)]
    children = [Optional('self', idname='adjunct', one2one=True), WavelengthHolder, CASU, System]
    SpectrumType = L1SuperstackSpectrum

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str], **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[int, 'HDU'], 'File', List[_BaseHDU]]:
        return super().read_hdus(directory, fname, casu=hierarchies['casu'], system=hierarchies['system'],
                                 l1single_files=hierarchies['l1single_files'],
                                 obspec=hierarchies['obspec'], arm_config=hierarchies['arm_config'],
                                 l1stack_spectra=hierarchies['l1superstack_spectra'], adjunct=hierarchies['adjunct'])

    @classmethod
    def fname_from_runid(cls, runid):
        return f'superstack_{runid:07.0f}.fit'


class L1SupertargetFile(L1StackedFile):
    singular_name = 'l1supertarget_file'
    match_pattern = 'WVE_.+\.fit'
    parents = [Multiple(L1SingleFile, maxnumber=10, constrain=(WeaveTarget, ArmConfig)),
               OneOf(L1SupertargetSpectrum, one2one=True)]
    children = [Optional('self', idname='adjunct', one2one=True), WavelengthHolder, CASU, System]
    SpectrumType = L1SupertargetSpectrum
    recommended_batchsize = None

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None, part=None):
        raise NotImplementedError


hierarchies = [i[-1] for i in inspect.getmembers(sys.modules[__name__], _predicate)]